"""
Make SMILES into Feature Matrix and generate a full dataset.

Featurizor:
    DChemPy.Featurization.MFFFeat: MFF Featurizor
    DChemPy.Featurization.MFFFeatWithFrags: MFF Featurizor with specific fragments
    DChemPy.Featurization.RDKitDescFeat: RDKit Descriptor Featurizor
    DChemPy.Featurization.RDKitDescFeatWithName: RDKit Descriptor Featurizor with specific feature name
    DChemPy.Featurization.ConjuMolFeat: Conjugation descriptor Featurizor
    DChemPy.Featurization.NormMolFeat: 'Conjugation descriptor' Featurizor for the whole molecule, not only conju-stru
    DChemPy.Featurization.PeptideFeat: Peptide's amino sequence index featurizor

Tools:
    DChemPy.Featurization.PeptideToSmi: Turn list(peptide) into list(SMILES)
    DChemPy.Featurization.CombineX: Combine a series of X matrix
    DChemPy.Featurization.CombineTitle: Combine a series of title
    DChemPy.Featurization.CheckDataset: Check datasets: delete rows with nan&inf; delete columns whose maximum==minimum
    DChemPy.Featurization.MakeSaveY: Save y, ln(y), y/MolWt and ln(y/MolWt)
    DChemPy.Featurization.SaveXSmiTitle: Save X, title and smiles

Example:
    from DChemPy.Featurization import *
    X_mff, title_mff = MFFFeat(smiles_list, radius=3)
    X_rdkit , title_rdkit = RDKitDescFeat(smiles_list)
    X_conju, title_conju = ConjuMolFeat(smiles_list, title_mff)
    X = CombineX([X_mff, X_rdkit, X_conju])
    title = CombineTitle([title_mff, title_rdkit, title_conju])
    X_out, y_out, title_out, smiles_out = CheckDataset(X, y, title, smiles_list)
    MakeSaveY(y_out, value_ln=True, directory='data_out', note='_True')
    DChemPy.Featurization.SaveXSmiTitle(X_out, smiles_out, title_out, directory='data_out')
"""


from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import Descriptors
from rdkit.Chem import rdMolDescriptors
from rdkit import DataStructs
from rdkit.Chem.EState.EState import EStateIndices
import numpy as np
import gc
import os
from pathlib import Path


def MFFFeat(smiles_rd, radius=2, addHatom=False, reserve_rate=0.05, exact_piece_reserve=False, exact_piece_smi=None,
            exact_piece_reserve_rate=0.0, turn_bool=False, turn_bool_mode='a', atom_count_control=True,
            min_frag_length=2, max_frag_length=50, omit_atom_count_control_frag_list=None, similarity_cal=False,
            similarity_method='Tanimoto', similarity_mode='mean'):
    """
    Generate MFF feature matrix.

    :param smiles_rd: List of SMILES string
    :param radius: ECFP Radius, default as '2'
    :param addHatom: Whether to add H in Mol, default as 'False'
    :param reserve_rate: Threshold of keeping a fragment column, default as '0.05'
    :param exact_piece_reserve: Whether to ignore 'reserve_rate' and keep fragments with specific smaller fragments,
                                default as 'False'
    :param exact_piece_smi: List of fragments to keep, default as 'None'
    :param exact_piece_reserve_rate:  Keep threshold of 'exact_piece_smi', default as '0.0'
    :param turn_bool: Whether to turn all MFF features into boolean form, default as 'False'
    :param turn_bool_mode: Mode of 'turn_bool', default as 'a', also can use 'b'
    :param atom_count_control: Whether to use atom-num control on fragments, default as 'True'
    :param min_frag_length: Minimum size of fragments, default as '2'
    :param max_frag_length: Maximum size of fragments, default as '50'
    :param omit_atom_count_control_frag_list: List of fragments to ignore 'atom_count_control', default as 'None'
    :param similarity_cal: Whether to calculate similarity with all other fragments for MFF values of 0,
                           default as 'False'
    :param similarity_method: Mode of similarity calculation, default as 'Tanimoto', also can use 'MACCS'
    :param similarity_mode: Calculate method of similarity calculation, default as 'mean', also can use 'max'

    :return: A tuple of (X, title)
             X: numpy array with shape of (len(smiles_rd), num_of_frags)
             title: numpy array with shape of (num_of_frags, 1)

    example:
    X_mff, title_mff = DChemPy.Featurization.MFFFeat(smiles_list)
    X_mff, title_mff = DChemPy.Featurization.MFFFeat(smiles_list, radius=3)
    """
    if type(smiles_rd) is np.ndarray:
        smiles_rd = smiles_rd.flatten().tolist()
    frag_smiles_list = []
    mol_list_temp = []
    for _ in range(len(smiles_rd)):
        smi = smiles_rd[_]
        mol = Chem.MolFromSmiles(smi)
        if addHatom:
            mol = Chem.AddHs(mol)
        mol_list_temp.append(mol)
        info = {}
        fp = rdMolDescriptors.GetMorganFingerprint(mol, radius, useChirality=True, useBondTypes=True, useFeatures=True,
                                                   bitInfo=info)
        fp = fp.GetNonzeroElements()
        for fragment_id, count in fp.items():
            root, radius = info[fragment_id][0]
            env = Chem.FindAtomEnvironmentOfRadiusN(mol, radius, root)
            frag = Chem.PathToSubmol(mol, env)
            f_smiles = Chem.MolToSmiles(frag)
            if f_smiles != '':
                frag_smiles_list.append(f_smiles)
    print('Number of all original fragments:', len(frag_smiles_list))
    title = np.unique(frag_smiles_list)
    data = np.zeros((len(smiles_rd), len(title)))
    for i in range(len(smiles_rd)):
        m = mol_list_temp[i]
        for j in range(len(title)):
            f = title[j]
            patt = Chem.MolFromSmarts(f)
            atomids = m.GetSubstructMatches(patt)
            data[i, j] = len(atomids)
    print('Before delete sparse columns: shape of data:', data.shape)
    del mol_list_temp
    gc.collect()
    # Delete less filled fragments:
    delete = []
    for i in range(data.shape[1]):
        filled_rate = sum(data[:, i] != 0) / data.shape[0]
        if filled_rate < reserve_rate:
            if exact_piece_reserve:
                if exact_piece_smi is None:
                    exact_piece_smi = []
                m = Chem.MolFromSmarts(title[i])
                for patt_temp in exact_piece_smi:
                    patt = Chem.MolFromSmarts(patt_temp)
                    flag = m.HasSubstructMatch(patt)
                    if flag:
                        break
                if filled_rate >= exact_piece_reserve_rate:
                    continue
            delete.append(i)
    data = np.delete(data, delete, axis=1)
    title = np.array(title).reshape(1, len(title))
    title = np.delete(title, delete, axis=1)
    title = list(title.flatten())
    print('After delete sparse columns: shape of data:', data.shape)
    if atom_count_control:
        if omit_atom_count_control_frag_list is None:
            omit_atom_count_control_frag_list = []
        delete = []
        for i in range(data.shape[1]):
            flag = False
            m = Chem.MolFromSmarts(title[i])
            for patt_temp in omit_atom_count_control_frag_list:
                patt = Chem.MolFromSmarts(patt_temp)
                flag = m.HasSubstructMatch(patt)
                if flag:
                    break
            if flag:
                continue
            atom_count = len(m.GetAtoms())
            if atom_count < min_frag_length or atom_count > max_frag_length:
                delete.append(i)
        data = np.delete(data, delete, axis=1)
        title = np.array(title).reshape(1, len(title))
        title = np.delete(title, delete, axis=1)
        title = list(title.flatten())
        print('After atom-count control: shape of data:', data.shape)
    if turn_bool:
        bool_patt_list = title
        for bool_patt in bool_patt_list:
            index_t = title.index(bool_patt)
            feature_max = int(max(data[:, index_t]))
            # 制作新的bool数据数组
            if turn_bool_mode == 'a':
                matrix_bool = np.zeros((data.shape[0], feature_max))
            else:
                matrix_bool = np.zeros((data.shape[0], feature_max + 1))
            for i in range(data.shape[0]):
                if turn_bool_mode == 'a':
                    if data[i, index_t] >= 1:
                        matrix_bool[i, int(data[i, index_t] - 1)] = 1
                else:
                    matrix_bool[i, int(data[i, index_t])] = 1
            data = np.delete(data, [index_t], axis=1)
            title = np.array(title).reshape(1, len(title))
            title = np.delete(title, [index_t], axis=1)
            title = list(title.flatten())
            new_title = []
            if turn_bool_mode == 'a':
                for i in range(feature_max):
                    new_title.append('With_' + str(i + 1) + '_' + bool_patt)
            else:
                for i in range(feature_max + 1):
                    new_title.append('With_' + str(i) + '_' + bool_patt)
            data = np.hstack((data, matrix_bool))
            title = [*title, *new_title]
        delete = []
        for i in range(data.shape[1]):
            if sum(data[:, i]) == 0:
                delete.append(i)
        data = np.delete(data, delete, axis=1)
        title = np.array(title).reshape(1, len(title))
        title = np.delete(title, delete, axis=1)
        title = list(title.flatten())
        print('title length(after turn to bool):', len(title))
        print('data shape(after turn to bool):', data.shape, '\n')
    if similarity_cal:
        data = data.astype(float)
        for i in range(data.shape[0]):
            emp = []
            emp_list = []
            fill = []
            fill_list = []
            for j in range(len(title)):
                if data[i, j] >= 1:
                    fill.append(title[j])
                    fill_list.append(j)
                elif data[i, j] == 0:
                    emp.append(title[j])
                    emp_list.append(j)
            for smi1_idx in emp_list:
                similarity = []
                smi1 = title[smi1_idx]
                fps1 = Chem.RDKFingerprint(Chem.MolFromSmarts(smi1))
                for smi2_idx in fill_list:
                    smi2 = title[smi2_idx]
                    fps2 = Chem.RDKFingerprint(Chem.MolFromSmarts(smi2))
                    sm = 0
                    if similarity_method == 'Tanimoto':
                        sm = DataStructs.FingerprintSimilarity(fps1, fps2)
                    elif similarity_method == 'MACCS':
                        sm = DataStructs.FingerprintSimilarity(fps1, fps2, metric=DataStructs.DiceSimilarity)
                    sm *= data[i, smi2_idx]
                    similarity.append(sm)
                sim = 0
                if similarity_mode == 'max':
                    sim = np.max(similarity)
                elif similarity_mode == 'mean':
                    sim = np.mean(similarity)
                data[i, smi1_idx] += sim
    x = data
    title = np.array(title).reshape(x.shape[1], 1)
    return x, title


def MFFFeatWithFrags(smiles_rd, frag_list):
    """
    Generate MFF feature matrix with given fragments.

    :param smiles_rd: List of SMILES string of molecules
    :param frag_list: List of SMILES string of fragments

    :return: A tuple of (X, title)
             X: numpy array with shape of (len(smiles_rd), num_of_frags)
             title: numpy array with shape of (num_of_frags, 1)

    example:
    X_mff, temp = DChemPy.Featurization.MFFFeatWithFrags(smiles_list, frag_list)
    """
    if type(smiles_rd) is np.ndarray:
        smiles_rd = list(smiles_rd.flatten())
    if type(frag_list) is np.ndarray:
        frag_list = list(frag_list.flatten())
    x = np.zeros((len(smiles_rd), len(frag_list)))
    for i in range(len(smiles_rd)):
        smi = smiles_rd[i]
        mol = Chem.MolFromSmiles(smi)
        for j in range(len(frag_list)):
            f = frag_list[j]
            patt = Chem.MolFromSmarts(f)
            atomids = mol.GetSubstructMatches(patt)
            x[i, j] = len(atomids)
    title = np.array(frag_list).reshape(x.shape[1], 1)
    return x, title


def RDKitDescFeat(smiles_rd):
    """
    Generate RDKit feature matrix

    :param smiles_rd: List of SMILES string of molecules

    :return: A tuple of (X, title)
             X: numpy array with shape of (len(smiles_rd), num_of_features)
             title: numpy array with shape of (num_of_features, 1)

    example:
    X_rdkit, title_rdkit = DChemPy.Featurization.RDKitDescFeat(smiles_list)
    """
    allowed = ['NOCount', 'VSA_EState4', 'NumHDonors', 'SlogP_VSA12', 'NumRadicalElectrons', 'SlogP_VSA4',
                          'Kappa2', 'Chi2n', 'PEOE_VSA3', 'PEOE_VSA7', 'PEOE_VSA4', 'Chi1', 'MolWt', 'SMR_VSA1',
                          'SlogP_VSA9', 'VSA_EState9', 'MaxAbsPartialCharge', 'NumSaturatedHeterocycles',
                          'MaxPartialCharge', 'VSA_EState1', 'PEOE_VSA6', 'EState_VSA11', 'SMR_VSA4', 'EState_VSA7',
                          'VSA_EState2', 'NHOHCount', 'SlogP_VSA10', 'SMR_VSA7', 'PEOE_VSA9', 'NumAliphaticRings',
                          'EState_VSA8', 'PEOE_VSA5', 'BertzCT', 'SlogP_VSA1', 'SlogP_VSA6', 'PEOE_VSA1', 'VSA_EState7',
                          'MinAbsPartialCharge', 'LabuteASA', 'SlogP_VSA2', 'EState_VSA4', 'MolMR', 'Kappa1',
                          'NumHAcceptors', 'EState_VSA9', 'MolLogP', 'NumAromaticHeterocycles', 'BalabanJ',
                          'FractionCSP3', 'SMR_VSA3', 'RingCount', 'NumSaturatedRings', 'PEOE_VSA2',
                          'MaxAbsEStateIndex', 'Kappa3', 'Chi3n', 'NumRotatableBonds', 'Chi4n', 'VSA_EState3',
                          'SMR_VSA8', 'MinPartialCharge', 'EState_VSA6', 'SMR_VSA9', 'PEOE_VSA13',
                          'NumValenceElectrons', 'MaxEStateIndex', 'SMR_VSA6', 'VSA_EState8', 'EState_VSA2',
                          'NumAromaticCarbocycles', 'SMR_VSA10', 'SlogP_VSA3', 'HallKierAlpha', 'PEOE_VSA14',
                          'HeavyAtomCount', 'VSA_EState10', 'SlogP_VSA11', 'ExactMolWt', 'MinAbsEStateIndex', 'TPSA',
                          'PEOE_VSA10', 'SMR_VSA2', 'Chi1v', 'Chi4v', 'PEOE_VSA8', 'EState_VSA5', 'Chi1n',
                          'VSA_EState5', 'SlogP_VSA7', 'HeavyAtomMolWt', 'MinEStateIndex', 'NumAliphaticHeterocycles',
                          'VSA_EState6', 'Chi0v', 'SlogP_VSA5', 'SMR_VSA5', 'Chi0', 'Chi2v', 'NumSaturatedCarbocycles',
                          'NumAromaticRings', 'Chi0n', 'PEOE_VSA12', 'Chi3v', 'NumAliphaticCarbocycles',
                          'EState_VSA10', 'EState_VSA3', 'EState_VSA1', 'NumHeteroatoms', 'SlogP_VSA8', 'PEOE_VSA11']
    if type(smiles_rd) is np.ndarray:
        smiles_rd = smiles_rd.flatten().tolist()
    descriptors = []
    desc_list = []
    for descriptor, function in Descriptors.descList:
        if descriptor in allowed:
            descriptors.append(descriptor)
            desc_list.append((descriptor, function))
    title = np.array(descriptors).reshape(len(descriptors), 1)
    data = []
    for i in range(len(smiles_rd)):
        mol = Chem.MolFromSmiles(smiles_rd[i])
        res = []
        for desc_name, f in desc_list:
            res.append(f(mol))
        data.append(res)
    data = np.array(data)
    print('RDKit Desc Data Shape:', data.shape)
    x = data
    return x, title


def RDKitDescFeatWithName(smiles_rd, desc_name_list):
    """
    Generate RDKit feature matrix with specific feature name

    :param smiles_rd: List of SMILES string of molecules
    :param desc_name_list: List of descriptor name of RDKit descriptors

    :return: A tuple of (X, title)
             X: numpy array with shape of (len(smiles_rd), num_of_features)
             title: numpy array with shape of (num_of_features, 1)

    example:
    X_rdkit, title_rdkit = DChemPy.Featurization.RDKitDescFeatWithName(smiles_list, desc_name_list)
    """
    if type(smiles_rd) is np.ndarray:
        smiles_rd = smiles_rd.flatten().tolist()
    descriptors = []
    desc_list = []
    for descriptor, function in Descriptors.descList:
        if descriptor in desc_name_list:
            descriptors.append(descriptor)
            desc_list.append((descriptor, function))
    title = np.array(descriptors).reshape(len(descriptors), 1)
    data = []
    for i in range(len(smiles_rd)):
        mol = Chem.MolFromSmiles(smiles_rd[i])
        res = []
        for desc_name, f in desc_list:
            res.append(f(mol))
        data.append(res)
    data = np.array(data)
    print('RDKit Desc Data Shape:', data.shape)
    x = data
    return x, title


def ConjuMolFeat(smiles_rd, mff_title_list=None, keep_type='max'):
    """
    Generate Conjugation descriptor matrix

    :param smiles_rd: List of SMILES string of molecules
    :param keep_type: Mode of calculation for MOE-like descriptors, default as 'max', also can use 'mean' or 'acc-mean'
    :param mff_title_list: List of MFF titles, needed for calculation, default as None

    :return: A tuple of (X, title)
             X: numpy array with shape of (len(smiles_rd), num_of_features)
             title: numpy array with shape of (num_of_features, 1)

    example:
    X_conju, title_conju = DChemPy.Featurization.ConjuMolFeat(smiles_list, mff_title_list=mff_title_list)
    """
    if type(smiles_rd) is np.ndarray:
        smiles_rd = list(smiles_rd.flatten())
    if type(mff_title_list) is np.ndarray:
        mff_title_list = list(mff_title_list.flatten())
    # For MOE-like VSA Descriptors:
    dij_m = np.zeros((54, 54))
    ri_m = np.zeros((54, 1))
    ri_m[6, 0] = 1.950  # C
    ri_m[7, 0] = 1.950  # N
    ri_m[8, 0] = 1.779  # O
    ri_m[9, 0] = 1.496  # F
    ri_m[15, 0] = 2.287  # P
    ri_m[16, 0] = 2.185  # S
    ri_m[17, 0] = 2.044  # Cl
    ri_m[35, 0] = 2.166  # Br
    ri_m[53, 0] = 2.358  # I
    dij_m[6, 35] = 1.970  # C-Br
    dij_m[35, 6] = 1.970
    dij_m[7, 35] = 1.840  # N-Br
    dij_m[35, 7] = 1.840
    dij_m[6, 6] = 1.540  # C-C
    dij_m[7, 7] = 1.450  # N-N
    dij_m[8, 8] = 1.470  # O-O
    dij_m[6, 17] = 1.800  # C-Cl
    dij_m[17, 6] = 1.800
    dij_m[6, 9] = 1.350  # C-F
    dij_m[9, 6] = 1.350
    dij_m[6, 53] = 2.120  # C-I
    dij_m[53, 6] = 2.120
    dij_m[6, 7] = 1.470  # C-N
    dij_m[7, 6] = 1.470
    dij_m[6, 8] = 1.430  # C-N
    dij_m[8, 6] = 1.430
    dij_m[6, 15] = 1.850  # C-P
    dij_m[15, 6] = 1.850
    dij_m[6, 16] = 1.810  # C-S
    dij_m[16, 6] = 1.810
    dij_m[7, 8] = 1.460  # N-O
    dij_m[8, 7] = 1.460
    dij_m[7, 15] = 1.600  # N-P
    dij_m[15, 7] = 1.600
    dij_m[7, 16] = 1.760  # N-S
    dij_m[16, 7] = 1.760
    dij_m[8, 15] = 1.570  # O-P
    dij_m[15, 8] = 1.570
    dij_m[8, 16] = 1.570  # O-S
    dij_m[16, 8] = 1.570

    def find_conju(mol, a_m):
        patt_list_d = ['C=C', 'C#C', 'C#N', 'C=O', 'C=S', 'C=N', 'N=N', '[N+]([O-])=O']
        patt_list_m = ['N', 'O', 'S', 'F', 'Cl', 'Br', 'I', 'P']
        ring_list = []
        f_a = []
        patt = Chem.MolFromSmarts('c')
        atomids = mol.GetSubstructMatches(patt)
        atoms = mol.GetAtoms()
        temp_list = []

        def find_ring(atom_id, found_atoms):
            nonlocal a_m, ring_list, f_a, atoms, temp_list
            c_flag = False
            c_list = np.argwhere(a_m[atom_id] == 1).flatten().tolist()
            for atom in c_list:
                if atom not in f_a:
                    a = atoms[atom]
                    if a.IsInRing() and str(a.GetHybridization()) != 'SP3':
                        found_atoms.append(atom)
                        f_a.append(atom)
                        find_ring(atom, found_atoms)
                        c_flag = True
            if not c_flag:
                temp_list.append(found_atoms)

        for atom in atomids:
            a = atom[0]
            if a not in f_a:
                find_ring(a, [])
            if len(temp_list) > 0:
                max_ring = temp_list[0]
                for l in temp_list:
                    if len(l) > len(max_ring):
                        max_ring = l
                ring_list.append(max_ring)
            temp_list = []
        for patt in patt_list_d:
            f = Chem.MolFromSmarts(patt)
            atomids = mol.GetSubstructMatches(f)
            if len(atomids) > 0:
                for pair in atomids:
                    n_l = []
                    flag_f = False
                    for a in pair:
                        if a in f_a:
                            flag_f = True
                            break
                        neighbors = atoms[a].GetNeighbors()
                        for na in neighbors:
                            n_l.append(na.GetIdx())
                    if flag_f:
                        continue
                    temp = []
                    temp_r_id = []
                    for n in n_l:
                        if atoms[n].GetAtomicNum() in [6, 7, 8]:
                            for i in range(len(ring_list)):
                                ring = ring_list[i]
                                if n in ring:
                                    temp.append(ring)
                                    temp_r_id.append(i)
                    if len(temp) == 1:
                        ring_list[temp_r_id[0]].append(pair[0])
                        ring_list[temp_r_id[0]].append(pair[1])
                        f_a.append(pair[0])
                        f_a.append(pair[1])
                    else:
                        t_r = []
                        for r in temp:
                            t_r += r
                        t_r.append(pair[0])
                        t_r.append(pair[1])
                        temp_r_id.sort()
                        temp_r_id = np.unique(temp_r_id)
                        for i in reversed(temp_r_id):
                            del ring_list[i]
                        ring_list.append(t_r)
                        f_a.append(pair[0])
                        f_a.append(pair[1])
        for patt in patt_list_m:
            f = Chem.MolFromSmarts(patt)
            atomids = mol.GetSubstructMatches(f)
            if len(atomids) > 0:
                for atom in atomids:
                    a = atom[0]
                    if a not in f_a:
                        neighbors = atoms[a].GetNeighbors()
                        n_l = []
                        for na in neighbors:
                            n_l.append(na.GetIdx())
                        temp = []
                        temp_r_id = []
                        for n in n_l:
                            for i in range(len(ring_list)):
                                ring = ring_list[i]
                                if (n in ring) and (i not in temp_r_id):
                                    temp.append(ring)
                                    temp_r_id.append(i)
                        if len(temp) == 1:
                            ring_list[temp_r_id[0]].append(a)
                            f_a.append(a)
                        else:
                            t_r = []
                            for r in temp:
                                t_r += r
                            t_r.append(a)
                            temp_r_id.sort()
                            for i in reversed(temp_r_id):
                                del ring_list[i]
                            if len(t_r) > 1:
                                ring_list.append(t_r)
                                f_a.append(a)
        for i in range(len(atoms)):
            if i not in f_a:
                aa = atoms[i]
                if aa.GetSymbol() != 'C' or str(aa.GetHybridization()) != 'SP2':
                    continue
                aa_n = aa.GetNeighbors()
                flag = False
                for aaa in aa_n:
                    if aaa.GetIdx() in f_a:
                        flag = True
                        break
                if flag:
                    a = i
                    neighbors = atoms[a].GetNeighbors()
                    n_l = []
                    for na in neighbors:
                        n_l.append(na.GetIdx())
                    temp = []
                    temp_r_id = []
                    for n in n_l:
                        for i in range(len(ring_list)):
                            ring = ring_list[i]
                            if (n in ring) and (i not in temp_r_id):
                                temp.append(ring)
                                temp_r_id.append(i)
                    if len(temp) == 1:
                        ring_list[temp_r_id[0]].append(a)
                        f_a.append(a)
                    else:
                        t_r = []
                        for r in temp:
                            t_r += r
                        t_r.append(a)
                        temp_r_id.sort()
                        for i in reversed(temp_r_id):
                            del ring_list[i]
                        if len(t_r) > 1:
                            ring_list.append(t_r)
                            f_a.append(a)
        if len(ring_list) > 1:
            flag = True
            while flag:
                t_temp = int(len(ring_list) * (len(ring_list) - 1) / 2)
                _temp = 0
                break_flag = False
                for i in range(len(ring_list) - 1):
                    for ji in range(len(ring_list) - i - 1):
                        r_1 = ring_list[i]
                        r_2 = ring_list[i + ji + 1]
                        if np.sum(a_m[r_1, :][:, r_2]) == 0:
                            _temp += 1
                        else:
                            for ki in r_2:
                                ring_list[i].append(ki)
                            ring_list[i] = np.unique(ring_list[i])
                            del ring_list[i + ji + 1]
                            break_flag = True
                            break
                    if break_flag:
                        break
                if _temp == t_temp:
                    flag = False
        for i in range(len(ring_list)):
            ring_list[i] = np.unique(ring_list[i]).flatten().tolist()
        return ring_list, f_a

    def find_elec_num(kind, _hyb):
        one_list = ['C']
        two_list = ['N', 'O', 'S', 'P', 'F', 'Cl', 'Br', 'I']
        if kind in one_list:
            return 1
        elif kind in two_list:
            if _hyb == 'SP' and kind in ['N', 'P', 'O']:
                return 1
            elif _hyb == 'SP2' and kind in ['N', 'O', 'S']:
                return 1
            else:
                return 2

    CONJU_TITLE = []
    CONJU_PRE = ['Apperant-Elec-Count', 'PEOE-Charge', 'EState-Indice', 'Atomic-LogP', 'Atomic-MR']
    CONJU_NUM = 5 + 7 + 16 * len(CONJU_PRE)
    data_conju = np.zeros((len(smiles_rd), CONJU_NUM))
    descList = []
    allowedDescriptors = ['MolWt']
    for descriptor, function in Descriptors.descList:
        if descriptor in allowedDescriptors:
            descList.append((descriptor, function))

    for _ in range(len(smiles_rd)):
        smi = smiles_rd[_]
        mol = Chem.MolFromSmiles(Chem.MolToSmiles(Chem.MolFromSmiles(smi)))
        atoms = mol.GetAtoms()
        a_m = Chem.rdmolops.GetAdjacencyMatrix(mol)
        d_m = Chem.rdmolops.GetDistanceMatrix(mol)
        res = find_conju(mol, a_m)
        ring_list = res[0]
        f_a = res[1]
        conju_size_list = [len(r) for r in ring_list]
        # Calculate apparent conjugation charges:
        app_elec = []
        for a in range(len(atoms)):
            if a in f_a:
                a_kind = atoms[a].GetSymbol()
                hyb = str(atoms[a].GetHybridization())
                app_elec.append(find_elec_num(a_kind, hyb))
            else:
                app_elec.append(0)
        # Calculate PEOE Charges:
        AllChem.ComputeGasteigerCharges(mol, nIter=25)
        peoe_charge = [mol.GetAtomWithIdx(i).GetDoubleProp('_GasteigerCharge') for i in range(mol.GetNumAtoms())]
        # Calculate EState Indices:
        estate_index = EStateIndices(mol)
        # Calculate atomic contribution of LogP and MR:
        contribs = rdMolDescriptors._CalcCrippenContribs(mol)
        logp = [contribs[i][0] for i in range(len(contribs))]
        mr = [contribs[i][1] for i in range(len(contribs))]

        atom_props = [app_elec, peoe_charge, estate_index, logp, mr]

        if _ == 0:
            CONJU_TITLE.append('Num of Conju-Stru (MFF-Conju)')  # 1.
            CONJU_TITLE.append('Num of Conju-All-Atoms (MFF-Conju)')  # 2.
            CONJU_TITLE.append('Atom Num Conju-All Ratio (MFF-Conju)')  # 3.
            CONJU_TITLE.append('AtomWt Conju-All Ratio (MFF-Conju)')  # 4.
            CONJU_TITLE.append('Full-Mol Wiener Index (MFF-Conju)')  # 5.
            CONJU_TITLE.append('Individual Conju-Atom Number (MFF-Conju)')  # 6.
            CONJU_TITLE.append('Conju-Part-Wt (MFF-Conju)')  # 7.
            CONJU_TITLE.append('Conju-AtomicWt (MFF-Conju)')  # 8.
            CONJU_TITLE.append('Max Conju-Distance (MFF-Conju)')  # 9.
            CONJU_TITLE.append('Conju-Branch Index (MFF-Conju)')  # 10.
            CONJU_TITLE.append('Conju-Stru Wiener Index (MFF-Conju)')  # 11.
            CONJU_TITLE.append('Conju-Stru-VSA (MFF-Conju)')  # 12.
        # 1. Number of conjugation structures:
        data_conju[_, 0] = len(ring_list)
        # 2. Total number of atoms in all conjugation structures:
        data_conju[_, 1] = len(f_a)
        # 3. Ratio of conjugated and non-conjugated atoms:
        data_conju[_, 2] = len(f_a) / len(atoms)
        # 4. Ratio of conjugated and non-conjugated fragment weights:
        rval = []
        for desc_name, function in descList:
            rval.append(function(mol))
        wt_list = []
        mwt_list = []
        for r in ring_list:
            tt = 0
            for a in r:
                tt += atoms[a].GetMass()
            wt_list.append(tt)
            mwt_list.append(tt / len(r))
        data_conju[_, 3] = sum(wt_list) / rval[0]
        # 5. Wiener Index:
        if '.' in smi:
            mol22 = Chem.MolFromSmiles(smi.split('.')[0])
            dm22 = Chem.rdmolops.GetDistanceMatrix(mol22)
            data_conju[_, 4] = np.sum(dm22) / (2 * dm22.shape[0] * (dm22.shape[0] - 1))
        else:
            data_conju[_, 4] = np.sum(d_m) / (2 * d_m.shape[0] * (d_m.shape[0] - 1))

        # Indenpent features of conjugation structures:
        conju_props = []
        # 6. Atom numbers of individual conjugation structures:
        size_l = []
        for r in ring_list:
            size_l.append(len(r))
        conju_props.append(size_l)
        # 7. Weights of individual conjugation structures:
        conju_props.append(wt_list)
        # 8. Mean atomic weights of individual conjugation structures:
        conju_props.append(mwt_list)
        # 9. Size of individual conjugation structures:
        conju_max_dis = []
        for r in ring_list:
            conju_max_dis.append(np.max(d_m[r, :][:, r]))
        conju_props.append(conju_max_dis)

        # 10. Branching index of individual conjugation structures:
        branch_l = []
        for i in range(len(wt_list)):
            branch_l.append(np.sum(a_m[ring_list[i], :][:, ring_list[i]]) / (2 * size_l[i]))
        conju_props.append(branch_l)

        # 11. Weiner index of individual conjugation structures:
        wi_l = []
        for r in ring_list:
            d_m_temp = d_m[r, :][:, r]
            wi_l.append(np.sum(d_m_temp) / (2 * d_m_temp.shape[0] * (d_m_temp.shape[0] - 1)))
        conju_props.append(wi_l)

        # 12. VSA of individual conjugation structures:
        conju_vsa_l = []
        for r in ring_list:
            vsa_t = 0
            for i in range(len(r)):
                vsa_tt = 0
                atom = atoms[r[i]]
                n_l = atom.GetNeighbors()
                aid_1 = atom.GetAtomicNum()
                ar_1 = ri_m[aid_1, 0]
                for j in range(len(n_l)):
                    aid_2 = n_l[j].GetAtomicNum()
                    ar_2 = ri_m[aid_2, 0]
                    dij_i = dij_m[aid_1, aid_2]
                    dij = min(max(abs(ar_1 - ar_2), dij_i), ar_1 + ar_2)
                    vsa_tt += (ar_2 ** 2 - (ar_1 - dij) ** 2) / dij
                vsa_t += 4 * np.pi * ar_1 ** 2 - np.pi * ar_1 * vsa_tt
            conju_vsa_l.append(vsa_t)
        conju_props.append(conju_vsa_l)

        # Atomic Descriptors:
        for __ in range(len(atom_props)):
            PRE = 'Conju-' + CONJU_PRE[__] + '-'
            END_P = ' (MFF-Conju)'
            atom_props_list = atom_props[__]
            if _ == 0:
                CONJU_TITLE.append(PRE + 'Sum' + END_P)  # 13.1.
                CONJU_TITLE.append(PRE + 'AtomicMean' + END_P)  # 13.2.
                CONJU_TITLE.append(PRE + 'Maximum' + END_P)  # 13.3.
                CONJU_TITLE.append(PRE + 'Minimum' + END_P)  # 13.4.
                CONJU_TITLE.append(PRE + 'Delta' + END_P)  # 13.5.
                CONJU_TITLE.append(PRE + 'Influence' + END_P)  # 13.6.
                CONJU_TITLE.append(PRE + 'AtomicInfluence' + END_P)  # 13.7.
                CONJU_TITLE.append(PRE + 'PositiveDisCoef' + END_P)  # 13.8.
                CONJU_TITLE.append(PRE + 'PairMean-PositiveDisCoef' + END_P)  # 13.9.
                CONJU_TITLE.append(PRE + 'NegativeDisCoef' + END_P)  # 13.10.
                CONJU_TITLE.append(PRE + 'PairMean-NegativeDisCoef' + END_P)  # 13.11.
                CONJU_TITLE.append(PRE + 'GradSum' + END_P)  # 13.12.
                CONJU_TITLE.append(PRE + 'PairMean-Grad' + END_P)  # 13.13.
                CONJU_TITLE.append(PRE + 'MaxMinDisRatio' + END_P)  # 13.14.
                # CONJU_TITLE.append(PRE+'AppearantU'+END_P)  # 13.15.
                CONJU_TITLE.append(PRE + 'LaplaceSum' + END_P)  # 13.16.
                CONJU_TITLE.append(PRE + 'PairMean-Laplace' + END_P)  # 13.17.
            # 13.1. Sum without fragment:
            x_count_l = []
            # 13.2. Atomic mean without fragments:
            x_atom_mean_l = []
            # 13.3. Maximum with fragments:
            x_max_l = []
            # 13.4. Minimum with fragments:
            x_min_l = []
            # 13.5. Max-Min with fragments:
            x_delta_l = []
            # 13.6. Influence:
            x_infl_l = []
            # 13.7. Atomic influence:
            x_atom_infl_l = []
            # 13.8. One order distance coefficient:
            x_pos_dis_coef_l = []
            # 13.9. Pair-mean one order distance coefficient:
            x_pair_mean_pos_dis_coef_l = []
            # 13.10. Negative one order distance coefficient:
            x_neg_dis_coef_l = []
            # 13.11. Pair-mean negative one order distance coefficient:
            x_pair_mean_neg_dis_coef_l = []
            # 13.12. Sum of one order gradient:
            x_sum_grad_l = []
            # 13.13. Pair-mean one order gradient:
            x_pair_mean_grad_l = []
            # 13.14. Max distance ratio of fragment pairs:
            x_dis_ratio_l = []
            # 13.15. Potential energy:
            # x_u_l = []
            # 13.16. Sum of two order gradient:
            x_sum_laplace_l = []
            # 13.17. Pair-mean two order gradient:
            x_pair_mean_laplace_l = []
            # Properties calculation:
            for r in ring_list:
                a_p_l = [atom_props_list[a] for a in r]
                x_count_l.append(sum(a_p_l))  # 1.
                x_atom_mean_l.append(sum(a_p_l) / len(r))  # 2.
                frag_x = []
                frag_atom_id = []
                for i in range(len(mff_title_list)):
                    patt = mff_title_list[i]
                    f = Chem.MolFromSmarts(patt)
                    atomids = mol.GetSubstructMatches(f)
                    if len(atomids) > 0:
                        for j in range(len(atomids)):
                            peoe_flag = True
                            for k in atomids[j]:
                                if k not in r:
                                    peoe_flag = False
                                    break
                            if peoe_flag:
                                frag_atom_id.append(atomids[j])
                                x_temp = 0
                                for k in atomids[j]:
                                    x_temp += atom_props_list[k]
                                frag_x.append(x_temp)
                for j in range(len(r)):
                    atom_id = r[j]
                    frag_x.append(atom_props_list[atom_id])
                    frag_atom_id.append([atom_id])
                x_max_l.append(max(frag_x))  # 3.
                x_min_l.append(min(frag_x))  # 4.
                x_delta_l.append(max(frag_x) - min(frag_x))  # 5.
                f_1 = [i for i in frag_atom_id[frag_x.index(min(frag_x))]]
                f_2 = [i for i in frag_atom_id[frag_x.index(max(frag_x))]]
                s = np.max(d_m[f_1, :][:, f_2]) / np.max(d_m[r, :][:, r])
                x_dis_ratio_l.append(s)  # 14.
                # Calculate the influence.
                temp_infl = 0
                for a in r:
                    temp_infl += atom_props_list[a] * atoms[a].GetDegree()
                x_infl_l.append(temp_infl)  # 6.
                x_atom_infl_l.append(temp_infl / len(r))  # 7.
                # Calculate the distance index and one or two order gradients.
                ttp = 0
                tti = 0
                grad_temp = 0
                laplace_temp = 0
                count = 0
                for i in range(len(r)):
                    for j in range(len(r) - i - 1):
                        a = r[i]
                        b = r[i + j + 1]
                        count += 1
                        a_e = atom_props_list[a]
                        b_e = atom_props_list[b]
                        ttp += a_e * b_e * d_m[a, b]
                        tti += a_e * b_e / d_m[a, b]
                        grad_temp += abs(a_e - b_e) / d_m[a, b]
                        laplace_temp += abs(a_e - b_e) / (d_m[a, b] ** 2)
                x_pos_dis_coef_l.append(ttp)  # 8.
                x_pair_mean_pos_dis_coef_l.append(ttp / count)  # 9.
                x_neg_dis_coef_l.append(tti)  # 10.
                x_pair_mean_neg_dis_coef_l.append(tti / count)  # 11.
                x_sum_grad_l.append(grad_temp)  # 12.
                x_pair_mean_grad_l.append(grad_temp / count)  # 13.
                x_sum_laplace_l.append(laplace_temp)  # 16.
                x_pair_mean_laplace_l.append(laplace_temp / count)  # 17.
            conju_props.append(x_count_l)
            conju_props.append(x_atom_mean_l)
            conju_props.append(x_max_l)
            conju_props.append(x_min_l)
            conju_props.append(x_delta_l)
            conju_props.append(x_infl_l)
            conju_props.append(x_atom_infl_l)
            conju_props.append(x_pos_dis_coef_l)
            conju_props.append(x_pair_mean_pos_dis_coef_l)
            conju_props.append(x_neg_dis_coef_l)
            conju_props.append(x_pair_mean_neg_dis_coef_l)
            conju_props.append(x_sum_grad_l)
            conju_props.append(x_pair_mean_grad_l)
            conju_props.append(x_dis_ratio_l)
            conju_props.append(x_sum_laplace_l)
            conju_props.append(x_pair_mean_laplace_l)
        for i in range(len(conju_props)):
            index = 5 + i
            if keep_type == 'max':
                data_conju[_, index] = conju_props[i][conju_size_list.index(max(conju_size_list))]
            elif keep_type == 'mean':
                data_conju[_, index] = np.mean(conju_props[i])
            elif keep_type == 'acc-mean':
                temp = 0
                for j in range(len(conju_size_list)):
                    temp += conju_size_list[j] * conju_props[i][j]
                data_conju[_, index] = temp / sum(conju_size_list)
    x = data_conju
    title = np.array(CONJU_TITLE).reshape(len(CONJU_TITLE), 1)
    return x, title


def NormMolFeat(smiles_rd, mff_title_list=None):
    """
    Generate 'Conjugation' descriptor matrix for the whole molecule, not only the conjugation structure.

    :param smiles_rd: List of SMILES string of molecules
    :param mff_title_list: List of MFF titles, needed for calculation, default as None

    :return: A tuple of (X, title)
             X: numpy array with shape of (len(smiles_rd), num_of_features)
             title: numpy array with shape of (num_of_features, 1)

    example:
    X_norm, title_norm = DChemPy.Featurization.NormMolFeat(smiles_list, mff_title_list=mff_title_list)
    """
    if type(smiles_rd) is np.ndarray:
        smiles_rd = list(smiles_rd.flatten())
    if type(mff_title_list) is np.ndarray:
        mff_title_list = list(mff_title_list.flatten())
    # For MOE-like VSA Descriptors:
    dij_m = np.zeros((54, 54))
    ri_m = np.zeros((54, 1))
    ri_m[6, 0] = 1.950  # C
    ri_m[7, 0] = 1.950  # N
    ri_m[8, 0] = 1.779  # O
    ri_m[9, 0] = 1.496  # F
    ri_m[15, 0] = 2.287  # P
    ri_m[16, 0] = 2.185  # S
    ri_m[17, 0] = 2.044  # Cl
    ri_m[35, 0] = 2.166  # Br
    ri_m[53, 0] = 2.358  # I
    dij_m[6, 35] = 1.970  # C-Br
    dij_m[35, 6] = 1.970
    dij_m[7, 35] = 1.840  # N-Br
    dij_m[35, 7] = 1.840
    dij_m[6, 6] = 1.540  # C-C
    dij_m[7, 7] = 1.450  # N-N
    dij_m[8, 8] = 1.470  # O-O
    dij_m[6, 17] = 1.800  # C-Cl
    dij_m[17, 6] = 1.800
    dij_m[6, 9] = 1.350  # C-F
    dij_m[9, 6] = 1.350
    dij_m[6, 53] = 2.120  # C-I
    dij_m[53, 6] = 2.120
    dij_m[6, 7] = 1.470  # C-N
    dij_m[7, 6] = 1.470
    dij_m[6, 8] = 1.430  # C-N
    dij_m[8, 6] = 1.430
    dij_m[6, 15] = 1.850  # C-P
    dij_m[15, 6] = 1.850
    dij_m[6, 16] = 1.810  # C-S
    dij_m[16, 6] = 1.810
    dij_m[7, 8] = 1.460  # N-O
    dij_m[8, 7] = 1.460
    dij_m[7, 15] = 1.600  # N-P
    dij_m[15, 7] = 1.600
    dij_m[7, 16] = 1.760  # N-S
    dij_m[16, 7] = 1.760
    dij_m[8, 15] = 1.570  # O-P
    dij_m[15, 8] = 1.570
    dij_m[8, 16] = 1.570  # O-S
    dij_m[16, 8] = 1.570

    def find_elec_num(kind, hyb_t):
        one_list = ['C']
        two_list = ['N', 'O', 'S', 'P', 'F', 'Cl', 'Br', 'I']
        if kind in one_list:
            return 1
        elif kind in two_list:
            if hyb_t == 'SP' and kind in ['N', 'P', 'O']:
                return 1
            elif hyb_t == 'SP2' and kind in ['N', 'O', 'S']:
                return 1
            else:
                return 2

    CONJU_TITLE = []
    CONJU_PRE = ['Apperant-Elec-Count', 'PEOE-Charge', 'EState-Indice', 'Atomic-LogP', 'Atomic-MR']
    CONJU_NUM = 1 + 5 + 16 * len(CONJU_PRE)
    data_conju = np.zeros((len(smiles_rd), CONJU_NUM))
    descList = []
    allowedDescriptors = ['MolWt']
    for descriptor, function in Descriptors.descList:
        if descriptor in allowedDescriptors:
            descList.append((descriptor, function))

    for _ in range(len(smiles_rd)):
        smi = smiles_rd[_]
        mol = Chem.MolFromSmiles(Chem.MolToSmiles(Chem.MolFromSmiles(smi)))
        atoms = mol.GetAtoms()
        a_m = Chem.rdmolops.GetAdjacencyMatrix(mol)
        d_m = Chem.rdmolops.GetDistanceMatrix(mol)
        ring_list = [[a.GetIdx() for a in atoms]]
        f_a = [a.GetIdx() for a in atoms]
        conju_size_list = [len(r) for r in ring_list]
        # Calculate apparent conjugation charges:
        app_elec = []
        for a in range(len(atoms)):
            if a in f_a:
                a_kind = atoms[a].GetSymbol()
                hyb = str(atoms[a].GetHybridization())
                app_elec.append(find_elec_num(a_kind, hyb))
            else:
                app_elec.append(0)
        # Calculate PEOE Charges:
        AllChem.ComputeGasteigerCharges(mol, nIter=25)
        peoe_charge = [mol.GetAtomWithIdx(i).GetDoubleProp('_GasteigerCharge') for i in range(mol.GetNumAtoms())]
        # Calculate EState Indices:
        estate_index = EStateIndices(mol)
        # Calculate atomic contribution of LogP and MR:
        contribs = rdMolDescriptors._CalcCrippenContribs(mol)
        logp = [contribs[i][0] for i in range(len(contribs))]
        mr = [contribs[i][1] for i in range(len(contribs))]

        atom_props = [app_elec, peoe_charge, estate_index, logp, mr]

        if _ == 0:
            CONJU_TITLE.append('Num of Conju-Stru (MFF-Conju)')  # 1.
            CONJU_TITLE.append('Num of Conju-All-Atoms (MFF-Conju)')  # 2.
            CONJU_TITLE.append('Atom Num Conju-All Ratio (MFF-Conju)')  # 3.
            CONJU_TITLE.append('AtomWt Conju-All Ratio (MFF-Conju)')  # 4.
            CONJU_TITLE.append('Full-Mol Wiener Index (MFF-Conju)')  # 5.
            CONJU_TITLE.append('Individual Conju-Atom Number (MFF-Conju)')  # 6.
            CONJU_TITLE.append('Conju-Part-Wt (MFF-Conju)')  # 7.
            CONJU_TITLE.append('Conju-AtomicWt (MFF-Conju)')  # 8.
            CONJU_TITLE.append('Max Conju-Distance (MFF-Conju)')  # 9.
            CONJU_TITLE.append('Conju-Branch Index (MFF-Conju)')  # 10.
            CONJU_TITLE.append('Conju-Stru Wiener Index (MFF-Conju)')  # 11.
            CONJU_TITLE.append('Conju-Stru-VSA (MFF-Conju)')  # 12.
        # 5. Wiener Index:
        if '.' in smi:
            mol22 = Chem.MolFromSmiles(smi.split('.')[0])
            dm22 = Chem.rdmolops.GetDistanceMatrix(mol22)
            data_conju[_, 0] = np.sum(dm22) / (2 * dm22.shape[0] * (dm22.shape[0] - 1))
        else:
            data_conju[_, 0] = np.sum(d_m) / (2 * d_m.shape[0] * (d_m.shape[0] - 1))

        # Indenpent features of conjugation structures:
        wt_list = []
        mwt_list = []
        for r in ring_list:
            tt = 0
            for a in r:
                tt += atoms[a].GetMass()
            wt_list.append(tt)
            mwt_list.append(tt / len(r))
        # 6. Atom numbers of individual conjugation structures:
        size_l = []
        for r in ring_list:
            size_l.append(len(r))
        data_conju[_, 1] = size_l[0]
        # 9. Size of individual conjugation structures:
        conju_max_dis = []
        for r in ring_list:
            conju_max_dis.append(np.max(d_m[r, :][:, r]))
        data_conju[_, 2] = conju_max_dis[0]

        # 10. Branching index of individual conjugation structures:
        branch_l = []
        for i in range(len(wt_list)):
            branch_l.append(np.sum(a_m[ring_list[i], :][:, ring_list[i]]) / (2 * size_l[i]))
        data_conju[_, 3] = branch_l[0]

        # 11. Weiner index of individual conjugation structures:
        wi_l = []
        for r in ring_list:
            d_m_temp = d_m[r, :][:, r]
            wi_l.append(np.sum(d_m_temp) / (2 * d_m_temp.shape[0] * (d_m_temp.shape[0] - 1)))
        data_conju[_, 4] = wi_l[0]

        # 12. VSA of individual conjugation structures:
        conju_vsa_l = []
        for r in ring_list:
            vsa_t = 0
            for i in range(len(r)):
                vsa_tt = 0
                atom = atoms[r[i]]
                n_l = atom.GetNeighbors()
                aid_1 = atom.GetAtomicNum()
                ar_1 = ri_m[aid_1, 0]
                for j in range(len(n_l)):
                    aid_2 = n_l[j].GetAtomicNum()
                    ar_2 = ri_m[aid_2, 0]
                    dij_i = dij_m[aid_1, aid_2]
                    dij = min(max(abs(ar_1 - ar_2), dij_i), ar_1 + ar_2)
                    vsa_tt += (ar_2 ** 2 - (ar_1 - dij) ** 2) / dij
                vsa_t += 4 * np.pi * ar_1 ** 2 - np.pi * ar_1 * vsa_tt
            conju_vsa_l.append(vsa_t)
        data_conju[_, 5] = conju_vsa_l[0]

        # Atomic Descriptors:
        for __ in range(len(atom_props)):
            PRE = 'Conju-' + CONJU_PRE[__] + '-'
            END_P = ' (MFF-Conju)'
            atom_props_list = atom_props[__]
            if _ == 0:
                CONJU_TITLE.append(PRE + 'Sum' + END_P)  # 13.1.
                CONJU_TITLE.append(PRE + 'AtomicMean' + END_P)  # 13.2.
                CONJU_TITLE.append(PRE + 'Maximum' + END_P)  # 13.3.
                CONJU_TITLE.append(PRE + 'Minimum' + END_P)  # 13.4.
                CONJU_TITLE.append(PRE + 'Delta' + END_P)  # 13.5.
                CONJU_TITLE.append(PRE + 'Influence' + END_P)  # 13.6.
                CONJU_TITLE.append(PRE + 'AtomicInfluence' + END_P)  # 13.7.
                CONJU_TITLE.append(PRE + 'PositiveDisCoef' + END_P)  # 13.8.
                CONJU_TITLE.append(PRE + 'PairMean-PositiveDisCoef' + END_P)  # 13.9.
                CONJU_TITLE.append(PRE + 'NegativeDisCoef' + END_P)  # 13.10.
                CONJU_TITLE.append(PRE + 'PairMean-NegativeDisCoef' + END_P)  # 13.11.
                CONJU_TITLE.append(PRE + 'GradSum' + END_P)  # 13.12.
                CONJU_TITLE.append(PRE + 'PairMean-Grad' + END_P)  # 13.13.
                CONJU_TITLE.append(PRE + 'MaxMinDisRatio' + END_P)  # 13.14.
                # CONJU_TITLE.append(PRE+'AppearantU'+END_P)  # 13.15.
                CONJU_TITLE.append(PRE + 'LaplaceSum' + END_P)  # 13.16.
                CONJU_TITLE.append(PRE + 'PairMean-Laplace' + END_P)  # 13.17.
            # 13.1. Sum without fragment:
            x_count_l = []
            # 13.2. Atomic mean without fragments:
            x_atom_mean_l = []
            # 13.3. Maximum with fragments:
            x_max_l = []
            # 13.4. Minimum with fragments:
            x_min_l = []
            # 13.5. Max-Min with fragments:
            x_delta_l = []
            # 13.6. Influence:
            x_infl_l = []
            # 13.7. Atomic influence:
            x_atom_infl_l = []
            # 13.8. One order distance coefficient:
            x_pos_dis_coef_l = []
            # 13.9. Pair-mean one order distance coefficient:
            x_pair_mean_pos_dis_coef_l = []
            # 13.10. Negative one order distance coefficient:
            x_neg_dis_coef_l = []
            # 13.11. Pair-mean negative one order distance coefficient:
            x_pair_mean_neg_dis_coef_l = []
            # 13.12. Sum of one order gradient:
            x_sum_grad_l = []
            # 13.13. Pair-mean one order gradient:
            x_pair_mean_grad_l = []
            # 13.14. Max distance ratio of fragment pairs:
            x_dis_ratio_l = []
            # 13.15. Potential energy:
            # x_u_l = []
            # 13.16. Sum of two order gradient:
            x_sum_laplace_l = []
            # 13.17. Pair-mean two order gradient:
            x_pair_mean_laplace_l = []
            # Properties calculation:
            for r in ring_list:
                a_p_l = [atom_props_list[a] for a in r]
                x_count_l.append(sum(a_p_l))  # 1.
                x_atom_mean_l.append(sum(a_p_l) / len(r))  # 2.
                frag_x = []
                frag_atom_id = []
                for i in range(len(mff_title_list)):
                    patt = mff_title_list[i]
                    f = Chem.MolFromSmarts(patt)
                    atomids = mol.GetSubstructMatches(f)
                    if len(atomids) > 0:
                        for j in range(len(atomids)):
                            peoe_flag = True
                            for k in atomids[j]:
                                if k not in r:
                                    peoe_flag = False
                                    break
                            if peoe_flag:
                                frag_atom_id.append(atomids[j])
                                x_temp = 0
                                for k in atomids[j]:
                                    x_temp += atom_props_list[k]
                                frag_x.append(x_temp)
                for j in range(len(r)):
                    atom_id = r[j]
                    frag_x.append(atom_props_list[atom_id])
                    frag_atom_id.append([atom_id])
                x_max_l.append(max(frag_x))  # 3.
                x_min_l.append(min(frag_x))  # 4.
                x_delta_l.append(max(frag_x) - min(frag_x))  # 5.
                f_1 = [i for i in frag_atom_id[frag_x.index(min(frag_x))]]
                f_2 = [i for i in frag_atom_id[frag_x.index(max(frag_x))]]
                s = np.max(d_m[f_1, :][:, f_2]) / np.max(d_m[r, :][:, r])
                x_dis_ratio_l.append(s)  # 14.
                # Calculate the influence.
                temp_infl = 0
                for a in r:
                    temp_infl += atom_props_list[a] * atoms[a].GetDegree()
                x_infl_l.append(temp_infl)  # 6.
                x_atom_infl_l.append(temp_infl / len(r))  # 7.
                # Calculate the distance index and one or two order gradients.
                ttp = 0
                tti = 0
                grad_temp = 0
                laplace_temp = 0
                count = 0
                for i in range(len(r)):
                    for j in range(len(r) - i - 1):
                        a = r[i]
                        b = r[i + j + 1]
                        count += 1
                        a_e = atom_props_list[a]
                        b_e = atom_props_list[b]
                        ttp += a_e * b_e * d_m[a, b]
                        tti += a_e * b_e / d_m[a, b]
                        grad_temp += abs(a_e - b_e) / d_m[a, b]
                        laplace_temp += abs(a_e - b_e) / (d_m[a, b] ** 2)
                x_pos_dis_coef_l.append(ttp)  # 8.
                x_pair_mean_pos_dis_coef_l.append(ttp / count)  # 9.
                x_neg_dis_coef_l.append(tti)  # 10.
                x_pair_mean_neg_dis_coef_l.append(tti / count)  # 11.
                x_sum_grad_l.append(grad_temp)  # 12.
                x_pair_mean_grad_l.append(grad_temp / count)  # 13.
                x_sum_laplace_l.append(laplace_temp)  # 16.
                x_pair_mean_laplace_l.append(laplace_temp / count)  # 17.
            data_conju[_, 6+16*__] = x_count_l[0]
            data_conju[_, 7+16*__] = x_atom_mean_l[0]
            data_conju[_, 8+16*__] = x_max_l[0]
            data_conju[_, 9+16*__] = x_min_l[0]
            data_conju[_, 10+16*__] = x_delta_l[0]
            data_conju[_, 11+16*__] = x_infl_l[0]
            data_conju[_, 12+16*__] = x_atom_infl_l[0]
            data_conju[_, 13+16*__] = x_pos_dis_coef_l[0]
            data_conju[_, 14+16*__] = x_pair_mean_pos_dis_coef_l[0]
            data_conju[_, 15+16*__] = x_neg_dis_coef_l[0]
            data_conju[_, 16+16*__] = x_pair_mean_neg_dis_coef_l[0]
            data_conju[_, 17+16*__] = x_sum_grad_l[0]
            data_conju[_, 18+16*__] = x_pair_mean_grad_l[0]
            data_conju[_, 19+16*__] = x_dis_ratio_l[0]
            data_conju[_, 20+16*__] = x_sum_laplace_l[0]
            data_conju[_, 21+16*__] = x_pair_mean_laplace_l[0]
    x = data_conju
    title = np.array(CONJU_TITLE).reshape(len(CONJU_TITLE), 1)
    return x, title


def PeptideFeat(pep_list):
    """
    Generate simple amino acid index for peptides
    All amino piece should be represented as numbers as below:
    ref = {'O': 0, 'G': 1, 'A': 2, 'V': 3, 'L': 4, 'I': 5, 'M': 6, 'C': 7, 'T': 8, 'S': 9,
           'N': 10, 'Q': 11, 'D': 12, 'E': 13, 'F': 14, 'Y': 15, 'W': 16, 'H': 17, 'P': 18, 'K': 19, 'R': 20}

    :param pep_list: List of peptide sequence, for example ['A.G.H.C.T', 'V.M.A.Q']

    :return: A tuple of (X, title)
             X: numpy array with shape of (len(pep_list), max_num_of_amino_pieces)
             title: numpy array with shape of (max_num_of_amino_pieces, 1)

    example:
    X_pep, title_pep = DChemPy.Featurization.PeptideFeat(pep_list)
    """
    if type(pep_list) is np.ndarray:
        pep_list = list(pep_list.flatten())
    ref = {'O': 0, 'G': 1, 'A': 2, 'V': 3, 'L': 4, 'I': 5, 'M': 6, 'C': 7, 'T': 8, 'S': 9,
           'N': 10, 'Q': 11, 'D': 12, 'E': 13, 'F': 14, 'Y': 15, 'W': 16, 'H': 17, 'P': 18, 'K': 19, 'R': 20}
    max_length = max([len(p) for p in pep_list])
    data = np.zeros((len(pep_list), max_length))
    for i in range(len(pep_list)):
        pep = pep_list[i]
        amino = pep.split('.')
        for j in range(len(amino)):
            data[i, j] = ref[amino[j]]
    x = data
    title = np.array(['A'+str(i+1) for i in range(max_length)]).reshape(max_length, 1)
    return x, title


def PeptideToSmi(pep_list):
    """
    Turn a list of peptide into a list of SMILES

    :param pep_list: List of peptide sequence, for example ['A.G.H.C.T', 'V.M.A.Q']

    :return: smiles_list

    example:
    smi_list = DChemPy.Featurization.PeptideToSmi(pep_list)
    """
    smi_list = []
    for pep in pep_list:
        s = 'PEPTIDE1{' + pep + '}$$$$'
        mol = Chem.MolFromHELM(s)
        smi = Chem.MolToSmiles(mol)
        smi_list.append(smi)
    smi_list = np.array(smi_list).reshape(len(smi_list), 1)
    return smi_list


def CombineX(X_list):
    """
    Connect a series of X in X_list

    :param X_list: list of X matrix, for example [X_mff, X_rdkit, X_conju]

    :return: X

    example:
    X_out =  DChemPy.Featurization.CombineX([X_mff, X_rdkit, X_conju])
    """
    X = X_list[0]
    if len(X_list) >= 2:
        for i in range(len(X_list) - 1):
            X = np.hstack((X, X_list[i + 1]))
    return X


def CombineTitle(title_list):
    """
    Connect a series of title in title_list

    :param title_list: list of title, for example [title_mff, title_rdkit, title_conju]

    :return: title
    note: output title matrix with shape of (total_num_features, 1)

    example:
    title_out =  DChemPy.Featurization.CombineTitle([title_mff, title_rdkit, title_conju])
    """
    for i in range(len(title_list)):
        if type(title_list[i]) is np.ndarray:
            title_list[i] = title_list[i].reshape(title_list[i].shape[0], 1)
        elif type(title_list[i]) is list:
            title_list[i] = np.array(title_list[i]).reshape(len(title_list[i]), 1)
    title_init = title_list[0]
    if len(title_list) >= 2:
        for i in range(len(title_list) - 1):
            title_init = np.vstack((title_init, title_list[i + 1]))
    return title_init


def CheckDataset(x, y, title, smiles=None):
    """
    Check the dataset: delete rows with nan & inf, delete columns with maximum==minimum

    :param x: Feature Matrix
    :param y: Label Values with type of numpy.array and shape of (num_samples, 1)
    :param title: Title with type of numpy.array and shape of (num_features, 1)
    :param smiles: SMILES array with type of numpy.array and shape of (num_samples, 1), optional, default as None

    :return: Tuple of (x, y, title) or (x, y, title, smiles)

    example:
    X_out, y_out, title_out = DChemPy.Featurization.CheckDataset(X, y, title)
    X_out, y_out, title_out, smiles_out = DChemPy.Featurization.CheckDataset(X, y, title, smiles)
    """
    if type(y) is np.ndarray:
        y = y.reshape(y.shape[0], 1)
    elif type(y) is list:
        y = np.array(y).reshape(len(y), 1)
    if type(title) is np.ndarray:
        title = title.reshape(title.shape[0], 1)
    elif type(y) is list:
        y = np.array(y).reshape(len(y), 1)
    if smiles is not None:
        if type(smiles) is np.ndarray:
            smiles = smiles.reshape(smiles.shape[0], 1)
        elif type(smiles) is list:
            smiles = np.array(smiles).reshape(len(smiles), 1)
    del_list_row = []
    for i in range(x.shape[0]):
        if (np.isnan(x[i, :].astype(float)).sum() > 0) or (np.inf in x[i, :].astype(float)):
            del_list_row.append(i)
            continue
    del_list_col = []
    for i in range(x.shape[1]):
        if max(x[:, i].astype(float)) == min(x[:, i].astype(float)):
            del_list_col.append(i)
            continue
    x = np.delete(x, del_list_row, axis=0)
    x = np.delete(x, del_list_col, axis=1)
    y = np.delete(y, del_list_row, axis=0)
    title = np.delete(title, del_list_col, axis=0)
    if smiles is None:
        return x, y, title
    else:
        smiles = np.delete(smiles, del_list_row, axis=0)
        return x, y, title, smiles


def MakeSaveY(y, value_ln=False, value_div_weight=False, smiles=None, directory=None, notes=''):
    """
    Turn Y into ln(y), y/MolWt, ln(y/MolWt)

    :param y: Label Values with type of numpy.array and shape of (num_samples, 1)
    :param value_ln: Whether to calculate ln(y), default as 'False'
    :param value_div_weight: Whether to calculate y/MolWt, default as 'False'.
                             If set to 'True', smiles should be input.
    :param smiles: List of SMILES, default as 'None'
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'
    :param notes: Notes in file name, default as '', for example '_True'

    :return: None

    example:
    DChemPy.Featurization.MakeSaveY(y, value_ln=True, note='_True')
    DChemPy.Featurization.MakeSaveY(y, value_ln=True, directory='data_out', note='_True')
    DChemPy.Featurization.MakeSaveY(y, value_ln=True, value_div_weight=True, smiles=smiles_list,
                                    directory='data_out', note='_True')
    """
    if type(y) is np.ndarray:
        y = y.reshape(y.shape[0], 1)
    elif type(y) is list:
        y = np.array(y).reshape(len(y), 1)
    if type(smiles) is np.ndarray:
        smiles = list(smiles.flatten())
    if value_div_weight:
        values_rd_dwt = []
        for descriptor, function in Descriptors.descList:
            if descriptor == 'MolWt':
                for i in range(len(smiles)):
                    smile = smiles[i]
                    mol = Chem.MolFromSmiles(str(smile))
                    v = y[i, 0] / function(mol)
                    values_rd_dwt.append(v)
        values_dwt_out = np.array(values_rd_dwt).reshape(len(values_rd_dwt), 1)
    if directory is None:
        np.savetxt('Values'+notes+'_'+str(y.shape[0])+'.csv', y, fmt='%s', delimiter=',')
        if value_ln:
            np.savetxt('Values' + notes + '_ln_' + str(y.shape[0]) + '.csv', np.log(y), fmt='%s', delimiter=',')
        if value_div_weight:
            np.savetxt('Values' + notes + '_wt_' + str(y.shape[0]) + '.csv', values_dwt_out,
                       fmt='%s', delimiter=',')
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, 'Values' + notes + '_' + str(y.shape[0]) + '.csv')
        np.savetxt(save_name, y, fmt='%s', delimiter=',')
        if value_ln:
            save_name = Path('', directory, 'Values' + notes + '_ln_' + str(y.shape[0]) + '.csv')
            np.savetxt(save_name, np.log(y), fmt='%s', delimiter=',')
        if value_div_weight:
            save_name = Path('', directory, 'Values' + notes + '_w_' + str(y.shape[0]) + '.csv')
            np.savetxt(save_name, values_dwt_out, fmt='%s', delimiter=',')


def SaveXSmiTitle(x, smiles, title, directory=None, notes=''):
    """
    Save X, Smiles and Title

    :param x: Feature Matrix
    :param smiles: SMILES with type of numpy.array and shape of (num_samples, 1)
    :param title: Title with type of numpy.array and shape of (num_featrues, 1)
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'
    :param notes: Notes in file name, default as ''

    :return: None

    example:
    DChemPy.Featurization.SaveXSmiTitle(X_out, smiles_out, title_out)
    DChemPy.Featurization.SaveXSmiTitle(X_out, smiles_out, title_out, directory='data_out')
    """
    if type(smiles) is np.ndarray:
        smiles = smiles.reshape(smiles.shape[0], 1)
    elif type(smiles) is list:
        smiles = np.array(smiles).reshape(len(smiles), 1)
    if type(title) is np.ndarray:
        title = title.reshape(title.shape[0], 1)
    elif type(title) is list:
        title = np.array(title).reshape(len(title), 1)
    if directory is None:
        np.savetxt('Features'+notes+'_'+str(x.shape[0])+'_'+str(x.shape[1])+'.csv', x, fmt='%s', delimiter=',')
        np.savetxt('Smiles'+notes+'_'+str(smiles.shape[0])+'.csv', smiles, fmt='%s', delimiter=',')
        np.savetxt('Titles'+notes+'_'+str(title.shape[0])+'.csv', x, fmt='%s', delimiter=',')
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, 'Features' + notes + '_' + str(x.shape[0]) + '_' + str(x.shape[1]) + '.csv')
        np.savetxt(save_name, x, fmt='%s', delimiter=',')
        save_name = Path('', directory, 'Smiles' + notes + '_' + str(smiles.shape[0]) + '.csv')
        np.savetxt(save_name, smiles, fmt='%s', delimiter=',')
        save_name = Path('', directory, 'Titles' + notes + '_' + str(title.shape[0]) + '.csv')
        np.savetxt(save_name, title, fmt='%s', delimiter=',')



"""
Basic Statistic Analysis

Contents:
DChemPy.Statistic.CovAnalysis: Covariance analysis of full feature matrix
DChemPy.Statistic.FeatRelation: Draw scatter plot of two features and calculate correlation coefficients
DChemPy.Statistic.CorAnalysis: Correlation analysis of full feature matrix
"""

import numpy as np
import scipy
import pandas as pd
import matplotlib.pyplot as plt
from pathlib import Path
import os


def CovAnalysis(X, y, title, draw_pair_num=0, directory=None, notes=''):
    """
    Covariance analysis of full feature matrix

    :param X: Feature matrix, numpy.array, shape (num_samples, num_features)
    :param y: Label values, numpy.array, shape (num_samples, )
    :param title: Title, numpy.array, shape (num_features, )
    :param draw_pair_num: Number of feature pairs with highest abs(cov), int, default as 0
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'
    :param notes: Notes in file name, default as '99'

    :return: None

    example:
    DChemPy.Statistic.CovAnalysis(X, y, title, draw_pair_num=10, directory='xxx')
    """
    if type(y) is np.ndarray:
        y = y.reshape(y.shape[0], )
    elif type(y) is list:
        y = np.array(y).reshape(len(y), )
    if type(title) is np.ndarray:
        title = title.reshape(title.shape[0], )
    elif type(title) is list:
        title = np.array(title).reshape(len(title), )
    for i in range(X.shape[1]):
        X[:, i] = (X[:, i] - np.mean(X[:, i])) / np.std(X[:, i])
    cov_m = np.cov(X, rowvar=False)
    if directory is None:
        save_name = Path('', 'Cov01' + notes + '_CovMatrix_' + str(cov_m.shape[1]) + '_Features.csv')
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, 'Cov01' + notes + '_CovMatrix_' + str(cov_m.shape[1]) + '_Features.csv')
    np.savetxt(fname=save_name, X=cov_m, fmt='%s', delimiter=',')
    fig = plt.figure(figsize=(10, 8), dpi=250)
    ax = fig.add_axes([0.05, 0.08, 0.91, 0.86])
    im = ax.imshow(cov_m, cmap='plasma_r', origin='lower')
    ax.set_xlabel('Features', fontsize=15)
    ax.set_ylabel('Features', fontsize=15)
    plt.suptitle('Covariance Matrix of ' + str(X.shape[1]) + ' Features', fontsize=20)
    cb = plt.colorbar(im)
    cb.set_label('Covariance', fontsize=15)
    if directory is None:
        save_name = Path('', 'Cov02' + notes + '_CovMatrix_' + str(cov_m.shape[1]) + '_Features.png')
    else:
        save_name = Path('', directory, 'Cov02' + notes + '_CovMatrix_' + str(cov_m.shape[1]) + '_Features.png')
    plt.savefig(save_name)
    data = []
    for i in range(cov_m.shape[0]):
        for j in range(cov_m.shape[1] - i - 1):
            data.append([i, j + 1 + i, cov_m[i, j + 1 + i]])
    data_m = np.array(data)
    fig = plt.figure(figsize=(10, 8), dpi=250)
    ax = fig.add_axes([0.09, 0.08, 0.88, 0.85])
    ax.hist(data_m[:, -1], bins=40, density=False, facecolor='#3CB371', edgecolor='#006400', alpha=0.9)
    ax.set_xlabel('Covariance', fontsize=17)
    ax.set_ylabel('Times', fontsize=17)
    plt.suptitle('Distribution of Covariance of ' + str(X.shape[1]) + ' Features', fontsize=20)
    if directory is None:
        save_name = Path('', 'Cov03' + notes + '_CovDis_' + str(cov_m.shape[1]) + '_Features.png')
    else:
        save_name = Path('', directory, 'Cov03' + notes + '_CovDis_' + str(cov_m.shape[1]) + '_Features.png')
    plt.savefig(save_name)
    plt.close()
    if draw_pair_num:
        permu = np.argsort(-np.abs(data_m[:, -1]))
        data_m_p = data_m[permu, :]
        for _ in range(draw_pair_num):
            x_label = title[int(data_m_p[_, 0]), ]
            y_label = title[int(data_m_p[_, 1]), ]
            if '/' in x_label:
                x_label = x_label.split('/')[0]
            if '/' in y_label:
                y_label = y_label.split('/')[0]
            c_v = data_m_p[_, 2]
            x_data = X[:, int(data_m_p[_, 0])].flatten().tolist()
            y_data = X[:, int(data_m_p[_, 1])].flatten().tolist()
            fig = plt.figure(figsize=(10, 8), dpi=250)
            ax = fig.add_axes([0.11, 0.08, 0.88, 0.82])
            sc = ax.scatter(x_data, y_data, alpha=0.4, c=y, cmap='plasma_r', marker='o')
            ax.set_xlabel(x_label, fontsize=17)
            ax.set_ylabel(y_label, fontsize=17)
            plt.suptitle('Distribution of Data of ' + x_label + '\n vs ' + y_label + '   Covariance: ' + str(
                round(c_v, 5)), fontsize=20)
            plt.colorbar(sc)
            if directory is None:
                save_name = Path('', 'Cov04' + notes + '_Top-' + str(_ + 1) + '_' + x_label + '-' + y_label +
                                 '_' + str(round(c_v, 4)) + '.png')
            else:
                save_name = Path('', directory, 'Cov04' + notes + '_Top-' + str(_ + 1) + '_' + x_label + '-' + y_label +
                                 '_' + str(round(c_v, 4)) + '.png')
            plt.savefig(save_name)
            plt.close()


def FeatRelation(X, y, title, T_1='', T_2='y', x_label=None, y_label=None,
                 T_1_LOG=False, T_2_LOG=False, directory=None, notes=''):
    """
    Draw scatter plot of two features and calculate correlation coefficients
    
    :param X: Feature matrix, numpy.array, shape (num_samples, num_features)
    :param y: Label values, numpy.array, shape (num_samples, )
    :param title: Title, numpy.array, shape (num_features, )
    :param T_1: Name of Feature 1, string
    :param T_2: Name of Feature 2, string, default as 'y'
    :param x_label: Label of Feature 1 on plot, default as None (None means same as T_1)
    :param y_label: Label of Feature 2 on plot, default as None (None means same as T_2)
    :param T_1_LOG: Whether use lg(X[:, title.index(T_1)])
    :param T_2_LOG: Whether use lg(X[:, title.index(T_2)])
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'
    :param notes: Notes in file name, default as ''
    
    :return: None
    
    example:
    DChemPy.Statistic.FeatRelation(X, y, title, T_1='XXX', T_2='y', directory='xxx')
    """
    if type(y) is np.ndarray:
        y = y.reshape(y.shape[0], )
    elif type(y) is list:
        y = np.array(y).reshape(len(y), )
    if type(title) is np.ndarray:
        title = title.reshape(title.shape[0], )
    elif type(title) is list:
        title = np.array(title).reshape(len(title), )
    if x_label is None:
        x_label = T_1
    if y_label is None:
        y_label = T_2
    t_l = title.flatten().tolist()
    t_1 = t_l.index(T_1)
    if T_2 != 'y':
        t_2 = t_l.index(T_2)
    if '/' in x_label:
        x_label = x_label.split('/')[0]
    if '/' in y_label:
        y_label = y_label.split('/')[0]
    xxx = X[:, t_1].flatten().tolist()
    if T_1_LOG:
        x_data = np.log(xxx).flatten().tolist()
        x_label = 'Log(' + x_label + ')'
    else:
        x_data = xxx
    if T_2 != 'y':
        yy = X[:, t_2].flatten().tolist()
        if T_2_LOG:
            y_data = np.log(yy).flatten().tolist()
            y_label = 'Log(' + y_label + ')'
        else:
            y_data = X[:, t_2].flatten().tolist()
    else:
        y_data = y.flatten().tolist()
    x_1 = pd.Series(x_data)
    x_2 = pd.Series(y_data)
    pearson = x_1.corr(x_2, method="pearson")
    spearman = x_1.corr(x_2, method="spearman")
    kendall = x_1.corr(x_2, method="kendall")
    m = np.vstack((np.array(x_data), np.array(y_data)))
    print(m.shape)
    m[0, :] /= np.std(m[0, :])
    m[1, :] /= np.std(m[1, :])
    co = np.cov(m)[0, 1]
    if directory is None:
        save_name = Path('', 'FeatRelation' + notes + '_' + x_label + '_' + y_label + '.png')
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, 'Relation' + notes + '_' + x_label + '_' + y_label + '.png')
    fig = plt.figure(figsize=(9.2, 8.6), dpi=250)
    ax = fig.add_axes([0.16, 0.10, 0.80, 0.68])
    sc = ax.scatter(x_data, y_data, alpha=0.5, c=y, cmap='magma', marker='o', zorder=5)
    ax.grid(zorder=1)
    ax.set_xlabel(x_label, fontsize=15)
    ax.set_ylabel(y_label, fontsize=15)
    plt.suptitle('Distribution of Data of \n' + x_label + '\n & \n' + y_label + '\nPearson:' + str(round(pearson, 5)) +
                 '      Spearman:' + str(round(spearman, 5)) + '\nKendall:' + str(
        round(kendall, 5)) + '      Covariance:' + str(round(co, 5)), fontsize=18)
    plt.colorbar(sc)
    plt.savefig(save_name)


def CorAnalysis(X, y, title, directory=None, notes=''):
    """
    Correlation analysis of full feature matrix
    
    :param X: Feature matrix, numpy.array, shape (num_samples, num_features)
    :param y: Label values, numpy.array, shape (num_samples, )
    :param title: Title, numpy.array, shape (num_features, )
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'
    :param notes: Notes in file name, default as ''
    
    :return: None
    
    example:
    DChemPy.Statistic.CorAnalysis(X, y, title, directory='xxx')
    """
    if type(y) is np.ndarray:
        y = y.reshape(y.shape[0], )
    elif type(y) is list:
        y = np.array(y).reshape(len(y), )
    if type(title) is np.ndarray:
        title = title.reshape(title.shape[0], )
    elif type(title) is list:
        title = np.array(title).reshape(len(title), )
    y_s = pd.Series(y)
    pearson = []
    spearman = []
    kendall = []
    for i in range(X.shape[1]):
        x_i = pd.Series(X[:, i].flatten().tolist())
        pearson.append(x_i.corr(y_s, method="pearson"))
        spearman.append(x_i.corr(y_s, method="spearman"))
        kendall.append(x_i.corr(y_s, method="kendall"))
    # Pearson Correlation Coefficient
    pearson = np.array(pearson).reshape(len(pearson), 1)
    m_1 = np.hstack((title, pearson))
    arg = np.argsort(-pearson, axis=0).flatten().tolist()
    m_1 = m_1[arg, :]
    t = np.array(['Feature_Name', 'Pearson_Correlation_Coefficient'])
    m_1 = np.vstack((t, m_1))
    if directory is None:
        save_name = Path('', 'Cor01' + notes + '_PearsonCC_' + str(m_1.shape[0] - 1) + '_Features.csv')
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, 'Cor01' + notes + '_PearsonCC_' + str(m_1.shape[0] - 1) + '_Features.csv')
    np.savetxt(save_name, m_1, fmt='%s', delimiter=',')
    # Speraman Correlation Coefficient
    spearman = np.array(spearman).reshape(len(spearman), 1)
    m_2 = np.hstack((title, spearman))
    arg = np.argsort(-spearman, axis=0).flatten().tolist()
    m_2 = m_2[arg, :]
    t = np.array(['Feature_Name', 'Spearman_Correlation_Coefficient'])
    m_2 = np.vstack((t, m_2))
    if directory is None:
        save_name = Path('', 'Cor02' + notes + '_SpearmanCC_' + str(m_1.shape[0] - 1) + '_Features.csv')
    else:
        save_name = Path('', directory, 'Cor02' + notes + '_SpearmanCC_' + str(m_1.shape[0] - 1) + '_Features.csv')
    np.savetxt(save_name, m_2, fmt='%s', delimiter=',')
    # Kendall Correlation Coefficient
    kendall = np.array(kendall).reshape(len(kendall), 1)
    m_3 = np.hstack((title, kendall))
    arg = np.argsort(-kendall, axis=0).flatten().tolist()
    m_3 = m_3[arg, :]
    t = np.array(['Feature_Name', 'Kendall_Correlation_Coefficient'])
    m_3 = np.vstack((t, m_3))
    if directory is None:
        save_name = Path('', 'Cor03' + notes + '_KendallCC_' + str(m_1.shape[0] - 1) + '_Features.csv')
    else:
        save_name = Path('', directory, 'Cor03' + notes + '_KendallCC_' + str(m_1.shape[0] - 1) + '_Features.csv')
    np.savetxt(save_name, m_3, fmt='%s', delimiter=',')



"""
Modify SMILES String

Contents:
    DChemPy.Chem.NormalizeSmiles: Change SMILES into RDKit's standard format.
    DChemPy.Chem.NormalizeSmilesList: Change SMILES List into RDKit's standard format.
"""

import numpy as np
from rdkit import Chem


def NormalizeSmiles(smi):
    """
    Change SMILES into RDKit's standard format.

    :param smi: SMILES String
    :return: SMILES String with RDKit's standard format
    """
    n_smi = Chem.MolToSmiles(Chem.MolFromSmiles(smi))
    print(n_smi)
    return n_smi


def NormalizeSmilesList(smi_list):
    """
    Change SMILES List into RDKit's standard format.

    :param smi_list: List of SMILES String

    :return: List of SMILES String with RDKit's standard format
    """
    if type(smi_list) is np.ndarray:
        smi_list = list(smi_list.flatten())
    smi_out = []
    for smi in smi_list:
        n_smi = Chem.MolToSmiles(Chem.MolFromSmiles(smi))
        smi_out.append(n_smi)
    return smi_out



"""
Drawing plots in machine learning

Contents:
    DChemPy.Drawing.TPScatterOneEpoch: Draw true-pred scatter plot in one training epoch
    DChemPy.Drawing.TPScatterAllEpochTwoSet: Draw true-pred scatter plot of all training epochs
        note: can be used to draw Train+Test-Pred vs True or Test-Pred vs True
    DChemPy.Drawing.PerformanceHist: Draw hist plot of model performance
    DChemPy.Drawing.FeatureImportanceHist: Draw hist plot of feature importance
    DChemPy.Drawing.SHAPScatter: Draw SHAP scatter, line and hist plots
    DChemPy.Drawing.ALEPlot: Draw ALE line and hist plots
    DChemPy.Drawing.ReduceDimScatter: Draw 2D Scatter plot using 2D data from t-SNE or PCA
"""


import numpy as np
import matplotlib.pyplot as plt
import os
from pathlib import Path


def TPScatterOneEpoch(true_train, pred_train, true_test, pred_test, notes='99', directory=None, plt_title='Regression'):
    """
    Draw true-pred scatter plot in one training epoch

    :param true_train: true label values of training set, numpy.array, shape: (num_train_samples, )
    :param pred_train: predicted values of training set, numpy.array, shape: (num_train_samples, )
    :param true_test: true label values of testing set, numpy.array, shape: (num_test_samples, )
    :param pred_test: predicted values of testing set, numpy.array, shape: (num_test_samples, )
    :param notes: Notes in file name, default as '99'
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'
    :param plt_title: Title of scatter plot, default as 'Regression'

    :return: None

    example:
    DChemPy.Drawing.TPScatterOneEpoch(true_train, pred_train, true_test, pred_test, directory='xxx')
    """
    from sklearn.metrics import mean_squared_error, r2_score, mean_absolute_error
    if type(true_train) is np.ndarray:
        true_train = true_train.reshape(true_train.shape[0], )
    elif type(true_train) is list:
        true_train = np.array(true_train).reshape(len(true_train), )
    if type(pred_train) is np.ndarray:
        pred_train = pred_train.reshape(pred_train.shape[0], )
    elif type(pred_train) is list:
        pred_train = np.array(pred_train).reshape(len(pred_train), )
    if type(true_test) is np.ndarray:
        true_test = true_test.reshape(true_test.shape[0], )
    elif type(true_test) is list:
        true_test = np.array(true_test).reshape(len(true_test), )
    if type(pred_test) is np.ndarray:
        pred_test = pred_test.reshape(pred_test.shape[0], )
    elif type(pred_test) is list:
        pred_test = np.array(pred_test).reshape(len(pred_test), )
    train_mse = mean_squared_error(true_train, pred_train)
    train_mae = mean_absolute_error(true_train, pred_train)
    train_r2 = r2_score(true_train, pred_train)
    test_mse = mean_squared_error(true_test, pred_test)
    test_mae = mean_absolute_error(true_test, pred_test)
    test_r2 = r2_score(true_test, pred_test)
    fig = plt.figure(figsize=(7, 7.8), dpi=300)
    ax = fig.add_axes([0.15, 0.09, 0.78, 0.74])
    ax.scatter(true_train, pred_train, s=25, alpha=0.75)
    ax.scatter(true_test, pred_test, s=25, alpha=0.75)
    inter = max(max(true_train) - min(true_train), max(true_test) - min(true_test))
    left_limit = min(min(true_test)-0.2*inter, min(true_train)-0.2*inter)
    right_limit = max(max(true_test)+0.2*inter, max(true_train)+0.2*inter)
    ax.plot([left_limit, right_limit], [left_limit, right_limit], 'r:')
    ax.plot([left_limit, right_limit], [left_limit+1, right_limit+1], 'y:')
    ax.plot([left_limit, right_limit], [left_limit-1, right_limit-1], 'y:')
    ax.legend(['Correct', 'Correct+1', 'Correct-1', 'Train', 'Test'], loc='upper left',
              shadow=True, fontsize=17)
    ax.set_xlabel('True Values', fontsize=18)
    ax.set_ylabel('Prediction Values', fontsize=18)
    plt.suptitle(plt_title+' SingleFit True-Predict Scatter' +
                 '\nTrain R^2: '+str(round(train_r2, 5))+'  Test R^2: '+str(round(test_r2, 5)) +
                 '\nTrain MSE: '+str(round(train_mse, 5))+'  Test MSE: '+str(round(test_mse, 5)) +
                 '\nTrain MAE: '+str(round(train_mae, 5))+'  Test MAE: '+str(round(test_mae, 5)), fontsize=18)
    save_name = plt_title+'_'+notes+'_R2-'+str(round(test_r2, 4))+'_Scatter.png'
    if directory is None:
        save_name = Path('', save_name)
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, save_name)
    plt.savefig(save_name)
    plt.close()


def TPScatterAllEpochTwoSet(true_y, mean_list, std_list, mean_mse, mean_mae, mean_r2, num_epoch,
                            plt_title='Reg', draw_type='MeanAllPred-True', notes='03', directory=None):
    """
    Draw true-pred scatter plot of all training epochs

    :param true_y: true label values of all samples, numpy.array, shape: (num_train_samples, )
    :param mean_list: Mean prediction values of all samples of all epochs, list, len = num_samples
    :param std_list: STD of prediction values of all samples of all epochs, list, len = num_samples
    :param mean_mse: Mean MSE of regressor of all epochs
    :param mean_mae: Mean MAE of regressor of all epochs
    :param mean_r2: Mean R^2 score of regressor of all epochs
    :param num_epoch: Number of epochs
    :param plt_title: Title of scatter plot, default as 'Reg', name of regressor is recommended
    :param draw_type: 'MeanAllPred-True' or 'MeanTestPred-True', default as 'MeanAllPred-True'
    :param notes: Notes in file name, default as '03'
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'

    :return: None

    example:
    DChemPy.Drawing.TPScatterAllEpochTwoSet(true_y=y, mean_list=mean_list, std_list=std_list,
                                            mean_mse=0.5, mean_mae=0.3, mean_r2=0.6, num_epoch=100,
                                            plt_title='LightGBM', draw_type='MeanAllPred-True', directory='xxx')
    DChemPy.Drawing.TPScatterAllEpochTwoSet(true_y=y, mean_list=mean_test_list, std_list=std_test_list,
                                            mean_mse=0.5, mean_mae=0.3, mean_r2=0.6, num_epoch=100,
                                            plt_title='LightGBM', draw_type='MeanTestPred-True', directory='xxx')
    """
    if type(true_y) is np.ndarray:
        true_y = true_y.reshape(true_y.shape[0], )
    elif type(true_y) is list:
        true_y = np.array(true_y).reshape(len(true_y), )
    if type(mean_list) is np.ndarray:
        mean_list = list(mean_list.flatten())
    if type(std_list) is np.ndarray:
        std_list = list(std_list.flatten())
    fig = plt.figure(figsize=(10, 8), dpi=300)
    ax = fig.add_axes([0.11, 0.08, 0.88, 0.815])
    sc = ax.scatter(true_y, mean_list, alpha=0.55, c=std_list, cmap='viridis', marker='o')
    inter = max(max(true_y) - min(true_y), max(mean_list) - min(mean_list))
    left_limit = min(min(true_y) - 0.2 * inter, min(mean_list) - 0.2 * inter)
    right_limit = max(max(true_y) + 0.2 * inter, max(mean_list) + 0.2 * inter)
    ax.plot([left_limit, right_limit], [left_limit, right_limit], color='#B22222', linestyle=':', linewidth='2')
    ax.plot([left_limit, right_limit], [left_limit + 1, right_limit + 1], color='#FFA500', linestyle=':', linewidth='2')
    ax.plot([left_limit, right_limit], [left_limit - 1, right_limit - 1], color='#FFA500', linestyle=':', linewidth='2')
    ax.legend(['Correct', 'Correct+1', 'Correct-1', 'Mean of Prediction'], loc='upper left', fontsize=17, shadow=True)
    ax.set_xlabel('True Values', fontsize=17)
    if draw_type == 'MeanAllPred-True':
        ax.set_ylabel('Mean Pred Values (2T)', fontsize=17)
    elif draw_type == 'MeanTestPred-True':
        ax.set_ylabel('Mean Pred Values (Test)', fontsize=17)
    plt.suptitle(plt_title+' Scatter of '+draw_type+' of ' + str(num_epoch) + ' Rounds\n' +
                 'Mean Pred:  MSE: ' + str(round(mean_mse, 4)) +
                 '  MAE: ' + str(round(mean_mae, 4)) +
                 '  R^2: ' + str(round(mean_r2, 4)), fontsize=21)
    cb = plt.colorbar(sc)
    cb.set_label('Standard Deviation of Prediction Values', fontsize=17)
    plt.grid(which='major', color='#D5D5D5', alpha=0.5)
    save_name = plt_title+'_'+notes+'a_'+draw_type+'_Scatter_MeanR2-'+str(round(mean_r2, 4))+'.png'
    if directory is None:
        save_name = Path('', save_name)
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, save_name)
    plt.savefig(save_name)
    plt.close()

    out = np.hstack((np.array(mean_list).reshape(len(mean_list), 1),
                     np.array(std_list).reshape(len(mean_list), 1)))
    out = np.hstack((np.array(true_y).reshape(len(mean_list), 1), out))
    out = np.hstack((np.linspace(0, len(mean_list) - 1, len(mean_list), dtype=int).reshape(len(mean_list), 1), out))
    out = np.vstack((np.array(['Index', 'True Values', 'Mean Pred Values',
                               'Mean Standard Deviation of Pred Values']).reshape(1, 4),
                     out))
    save_name = plt_title+'_'+notes+'a_'+draw_type+'_RawData_MeanR2-'+str(round(mean_r2, 4))+'.csv'
    if directory is None:
        save_name = Path('', save_name)
    else:
        save_name = Path('', directory, save_name)
    np.savetxt(save_name, out, fmt='%s', delimiter=',')

    me_plot_list = []
    for i in range(len(mean_list)):
        me_plot_list.append(mean_list[i] - true_y[i])
    fig = plt.figure(figsize=(12, 6), dpi=300)
    ax = fig.add_axes([0.08, 0.11, 0.95, 0.815])
    x_idx = np.linspace(0, len(me_plot_list) - 1, len(me_plot_list))
    sc = ax.scatter(x_idx, me_plot_list, alpha=1.0, c=std_list, cmap='viridis', marker='o')
    ax.set_xlabel('Sample ID', fontsize=17)
    if draw_type == 'MeanAllPred-True':
        ax.set_ylabel('Deviation of Mean Pred Values (2T)', fontsize=17)
    elif draw_type == 'MeanTestPred-True':
        ax.set_ylabel('Deviation of Mean Pred Values (Test)', fontsize=17)
    ax.plot([0, len(me_plot_list)], [0, 0], color='#B22222', linestyle=':', linewidth='2')
    ax.plot([0, len(me_plot_list)], [1, 1], color='#FFA500', linestyle=':', linewidth='2')
    ax.plot([0, len(me_plot_list)], [-1, -1], color='#FFA500', linestyle=':', linewidth='2')
    cb = plt.colorbar(sc)
    cb.set_label('Standard Deviation of Prediction Values', fontsize=17)
    plt.grid(which='major', color='#D5D5D5', alpha=0.5)
    plt.suptitle(plt_title+' Scatter of '+draw_type+' Deviation vs ID of ' + str(num_epoch) + ' Rounds', fontsize=21)
    save_name = plt_title+'_'+notes+'a_'+draw_type+'_Deviation_MeanR2-'+str(round(mean_r2, 4))+'.png'
    if directory is None:
        save_name = Path('', save_name)
    else:
        save_name = Path('', directory, save_name)
    plt.savefig(save_name)
    plt.close()


def PerformanceHist(p_list, num_epoch, plt_title='Reg', n_bins=30, draw_type='MSE', notes='05', directory=None):
    """
    Draw hist plot of model performance

    :param p_list: List of model performance, list, len = num_epoch
    :param plt_title: Title of scatter plot, default as 'Reg', name of regressor is recommended
    :param num_epoch: Number of epochs
    :param n_bins: Number of bins in hist plot
    :param draw_type: 'MSE' or 'MAE' or 'R^2', default as 'MSE'
    :param notes: Notes in file name, default as '05'
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'

    :return: None

    example:
    DChemPy.Drawing.PerformanceHist(mae_list, 100, plt_title='LightGBM', draw_type='MAE', directory='xxx')
    """
    from scipy.stats import norm
    if type(p_list) is np.ndarray:
        p_list = list(p_list.flatten())
    mu_p = np.mean(p_list)
    sigma_p = np.std(p_list)
    p_array = np.array(p_list).reshape(len(p_list), 1)
    p_sorted = np.sort(p_array, axis=0)
    fig = plt.figure(figsize=(8, 8), dpi=300)
    ax = fig.add_axes([0.11, 0.08, 0.85, 0.81])
    if draw_type == 'MSE':
        fc = '#4682B4'
        ec = '#505050'
        lc = '#483D8B'
    elif draw_type == 'MAE':
        fc = '#3CB371'
        ec = '#006400'
        lc = '#B2F200'
    else:
        fc = '#FF6347'
        ec = '#FF4500'
        lc = '#B22222'
    n, bins_p, patches = ax.hist(p_sorted, bins=n_bins, density=1,
                                 facecolor=fc, edgecolor=ec, alpha=0.75, linewidth=1.6)
    pn = norm.pdf(bins_p, mu_p, sigma_p)
    plt.plot(bins_p, pn, color=lc, linestyle=':', linewidth='2')
    plt.suptitle('Distribution of '+plt_title+' '+draw_type+' of ' + str(num_epoch) +
                 ' Rounds\nMean:' + str(round(mu_p, 4)) +
                 '    STD:' + str(round(sigma_p, 4)), fontsize=22)
    ax.set_ylabel('Counts', fontsize=17)
    ax.set_xlabel(draw_type, fontsize=17)
    save_name = plt_title+'_'+notes+'_'+draw_type+'_Distribution.png'
    if directory is None:
        save_name = Path('', save_name)
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, save_name)
    plt.savefig(save_name)


def FeatureImportanceHist(f_i, title, plt_title='Reg', num_draw=10, notes='06', directory=None):
    """
    Draw hist plot of feature importance

    :param f_i: feature importance, numpy.array, shape (num_features, 1)
    :param title: title of all features, numpy.array, shape (num_featrues, )
    :param plt_title: Title of scatter plot, default as 'Reg', name of regressor is recommended
    :param num_draw: number of features put in the hist plot, default as 10
    :param notes: Notes in file name, default as '06'
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'

    :return: None

    example:
    DChemPy.Drawing.FeatureImportanceHist(f_i, title, plt_title='LightGBM', directory='xxx')
    """
    if type(f_i) is np.ndarray:
        f_i = f_i.reshape(f_i.shape[0], 1)
    elif type(f_i) is list:
        f_i = np.array(f_i).reshape(len(f_i), 1)
    if type(title) is np.ndarray:
        title = title.reshape(title.shape[0], )
    elif type(title) is list:
        title = np.array(title).reshape(len(title), )
    f_i[:, 0] = 100.0 * (f_i[:, 0] / (max(f_i[:, 0]) - min(f_i[:, 0])))
    sorted_idx = np.argsort(-f_i[:, 0])
    pos = np.arange(sorted_idx.shape[0]) + .5
    if len(pos) >= num_draw:
        fig = plt.figure(figsize=(11, 8), dpi=250)
        ax = fig.add_axes([0.11, 0.08, 0.85, 0.84])
        ax.barh(pos[:num_draw, ], f_i[sorted_idx[:num_draw, ], 0].flatten().tolist(),
                align='center', color='#CD5C5C', edgecolor='#A52A2A', linewidth=1.6)
        plt.yticks(pos[:num_draw, ], title[sorted_idx[:num_draw, ]])
        ax.set_xlabel('Relative Importance', fontsize=17)
        plt.suptitle('Feature Importance (First 10) of '+plt_title, fontsize=20)
        save_name = plt_title+'_'+notes+'a_Feature_Importance.png'
        if directory is None:
            save_name = Path('', save_name)
        else:
            if not os.path.exists(directory):
                os.mkdir(directory)
            save_name = Path('', directory, save_name)
        plt.savefig(save_name)
    else:
        num_f = sorted_idx.shape[0]
        fig = plt.figure(figsize=(11, 8), dpi=250)
        ax = fig.add_axes([0.11, 0.08, 0.85, 0.84])
        ax.barh(pos[:num_f, ], f_i[sorted_idx[:num_f, ], 0].flatten().tolist(),
                align='center', color='#CD5C5C', edgecolor='#A52A2A', linewidth=1.6)
        plt.yticks(pos[:num_f, ], title[sorted_idx[:num_f, ]])
        ax.set_xlabel('Relative Importance', fontsize=17)
        plt.suptitle('Feature Importance (First ' + str(int(num_f)) + ') of '+plt_title, fontsize=20)
        save_name = plt_title+'_'+notes+'a_Feature_Importance.png'
        if directory is None:
            save_name = Path('', save_name)
        else:
            if not os.path.exists(directory):
                os.mkdir(directory)
            save_name = Path('', directory, save_name)
        plt.savefig(save_name)
    feature_out = np.hstack((title.reshape(title.shape[0], 1), f_i))
    save_name = plt_title+'_'+notes+'b_All_Feature_Importance.csv'
    if directory is None:
        save_name = Path('', save_name)
    else:
        save_name = Path('', directory, save_name)
    np.savetxt(save_name, feature_out[sorted_idx.flatten().tolist(), :], fmt='%s', delimiter=',')


def SHAPScatter(feat_value, heter_value, label_value, f_name, norm_f_i=None, plt_title='Reg',
                inter_1d=20, notes='', directory=None):
    """
    Draw SHAP scatter, line and hist plots

    :param feat_value: Feature values, numpy.array, shape (num_samples, )
    :param heter_value: SHAP values, numpy.array, shape (num_samples, )
    :param label_value: Label values, numpy.array, shape (num_samples, )
    :param f_name: Feature name, string
    :param norm_f_i: Normalized feature importance, float or None, default as None
    :param plt_title: Title of scatter plot, default as 'Reg', name of regressor is recommended
    :param inter_1d: Number of segments when plotting SHAP Curve, default as 20
    :param notes: Notes in file name, default as ''
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'

    :return: None

    example:
    DChemPy.Drawing.SHAPScatter(feat_value=X[:, id], heter_value=shap_m[:, id], label_value=y,
                                f_name='XXX', plt_title='XGB', directory='xxx')
    """
    from math import floor
    if type(feat_value) is np.ndarray:
        feat_value = feat_value.reshape(feat_value.shape[0], )
    elif type(feat_value) is list:
        feat_value = np.array(feat_value).reshape(len(feat_value), )
    if type(heter_value) is np.ndarray:
        heter_value = heter_value.reshape(heter_value.shape[0], )
    elif type(heter_value) is list:
        heter_value = np.array(heter_value).reshape(len(heter_value), )
    if type(label_value) is np.ndarray:
        label_value = label_value.reshape(label_value.shape[0], )
    elif type(label_value) is list:
        label_value = np.array(label_value).reshape(len(label_value), )
    if f_name[-1] == ')':
        for j in range(len(f_name)):
            index = len(f_name) - j - 1
            if f_name[index] == '(':
                break
        f_name = f_name[:index - 1]
    if max(feat_value) > 0:
        ii = (max(feat_value) * 1.001 - min(feat_value)) / inter_1d
    else:
        ii = (max(feat_value) * 0.999 - min(feat_value)) / inter_1d
    fig = plt.figure(figsize=(8, 8), dpi=300)
    ax = fig.add_axes([0.11, 0.29, 0.70, 0.60])
    ax2 = fig.add_axes([0.11, 0.11, 0.70, 0.17])
    ax_bar = fig.add_axes([0.85, 0.11, 0.04, 0.78])
    x_idx = []
    x_inter = [min(feat_value)]
    x_ticks_blank = ['']
    x_ticks = [str(round(min(feat_value), 2))]
    y_mean_data = np.zeros((inter_1d, 1))
    y_mean_count = np.zeros((inter_1d, 1))
    for j in range(feat_value.shape[0]):
        y_mean_data[floor((feat_value[j, ] - min(feat_value)) / ii), 0] += heter_value[j, ]
        y_mean_count[floor((feat_value[j, ] - min(feat_value)) / ii), 0] += 1
    for j in range(inter_1d):
        x_idx.append(ii * (0.5 + j) + min(feat_value))
        x_inter.append(ii * j + ii + min(feat_value))
        x_ticks.append(str(round(ii * (j + 1) + min(feat_value), 2)))
        x_ticks_blank.append('')
        if y_mean_count[j, 0] != 0:
            y_mean_data[j, 0] /= y_mean_count[j, 0]
        else:
            y_mean_data[j, 0] = y_mean_data[j - 1, 0]
    x_data = feat_value.flatten().tolist()
    y_data = heter_value.flatten().tolist()
    y = label_value.flatten().tolist()
    sc = ax.scatter(x_data, y_data, c=y, cmap='viridis', marker='o', zorder=5)
    ax.plot(x_idx, y_mean_data, color='#EE82EE', linewidth=3, zorder=10, marker='o')
    ax.set_ylim(min(y_data) - 0.2 * (max(y_data) - min(y_data)), max(y_data) + 0.1 * (max(y_data) - min(y_data)))
    t = []
    for j in range(feat_value.shape[0]):
        t.append(min(y_data) - 0.1 * (max(y_data) - min(y_data)))
    ax.scatter(x_data, t, marker='|', alpha=0.1, color='#C0C0C0')
    if norm_f_i is None:
        plt.suptitle('SHAP Value vs Feature Value Plot of\n' + f_name +
                     '  FI-abs(SHAP):' + str(round(abs(sum(y_data)), 2)),
                     fontsize=22)
    else:
        plt.suptitle('SHAP Value vs Feature Value Plot of\n' + f_name + '  NormFI:' + str(round(norm_f_i, 2)),
                     fontsize=22)
    ax2.hist(x_data, inter_1d, density=1, facecolor='#3CB371', edgecolor='#006400', alpha=0.75, linewidth=1.6)
    ax2.set_xlabel(f_name, fontsize=17)
    ax.set_ylabel('SHAP Value of y', fontsize=17)
    ax2.set_ylabel('Count', fontsize=17)
    ax.set_xticks(x_inter)
    ax.set_xticklabels(x_ticks_blank)
    ax2.set_xticks(x_inter)
    ax2.set_xticklabels(x_ticks, rotation=45, fontsize=11)
    ax.grid(which='major', color='#D5D5D5', alpha=0.5, zorder=2)
    cb = plt.colorbar(sc, cax=ax_bar)
    cb.set_label('Label Value (Y)', fontsize=17)
    save_name = plt_title+'_'+notes+'SHAP-Scatter_' + f_name + '.png'
    if directory is None:
        save_name = Path('', save_name)
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, save_name)
    plt.savefig(save_name)
    plt.close()

    o = np.hstack((feat_value.reshape(feat_value, 1), heter_value.reshape(heter_value.shape[0], 1)))
    o = np.vstack((np.array(['Feature Values', 'SHAP values']).reshape(1, 2), o))
    save_name = plt_title + '_' + notes + 'b_SHAP-Scatter_Data_' + f_name + '.csv'
    if directory is None:
        save_name = Path('', save_name)
    else:
        save_name = Path('', directory, save_name)
    np.savetxt(save_name, o, fmt='%s', delimiter=',')


def ALEPlot(feat_value, heter_value, f_name, plt_title='Reg', inter_1d=20, notes='', directory=None):
    """
    Draw ALE line and hist plots

    :param feat_value: Feature values, numpy.array, shape (num_samples, )
    :param heter_value: ALE values, numpy.array, shape (num_samples, )
    :param f_name: Feature name, string
    :param plt_title: Title of line plot, default as 'Reg', name of regressor is recommended
    :param inter_1d: Number of segments when plotting ALE Curve, default as 20
    :param notes: Notes in file name, default as ''
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'

    :return: None

    example:
    DChemPy.Drawing.ALEPlot(feat_value=X[:, id], heter_value=ale_data, f_name='XXX', plt_title='XGB', directory='xxx')
    """
    if type(feat_value) is np.ndarray:
        feat_value = feat_value.reshape(feat_value.shape[0], )
    elif type(feat_value) is list:
        feat_value = np.array(feat_value).reshape(len(feat_value), )
    if type(heter_value) is np.ndarray:
        heter_value = heter_value.reshape(heter_value.shape[0], )
    elif type(heter_value) is list:
        heter_value = np.array(heter_value).reshape(len(heter_value), )
    fig = plt.figure(figsize=(8, 8), dpi=300)
    ax = fig.add_axes([0.11, 0.29, 0.85, 0.64])
    ax2 = fig.add_axes([0.11, 0.11, 0.85, 0.17])
    x_idx = []
    x_inter = [min(feat_value)]
    x_ticks_blank = ['']
    x_ticks = [str(round(min(feat_value), 2))]
    ii = (max(feat_value) * 1.00001 - min(feat_value)) / inter_1d
    for i in range(inter_1d):
        x_idx.append(ii * (0.5 + i) + min(feat_value))
        x_inter.append(ii * i + ii + min(feat_value))
        x_ticks.append(str(round(ii * (i + 1) + min(feat_value), 2)))
        x_ticks_blank.append('')
    y_data = heter_value.flatten().tolist()
    ax.plot(x_idx, y_data, color='#483D8B', linestyle=':', linewidth=3, marker='o',
            markersize=12, zorder=10)
    ax.set_ylim(min(y_data) - 0.2 * (max(y_data) - min(y_data)),
                max(y_data) + 0.1 * (max(y_data) - min(y_data)))
    t = []
    for i in range(feat_value.shape[0]):
        t.append(min(y_data) - 0.1 * (max(y_data) - min(y_data)))
    ax.scatter(feat_value, t, marker='|', alpha=0.1, color='#C0C0C0')
    plt.suptitle('1D ALE Plot of ' + f_name, fontsize=23)
    ax2.hist(feat_value, inter_1d, density=1, facecolor='#3CB371', edgecolor='#006400',
             alpha=0.75, linewidth=1.6)
    ax2.set_xlabel(f_name, fontsize=17)
    ax.set_ylabel('ALE Value of y', fontsize=17)
    ax2.set_ylabel('Count', fontsize=17)
    ax.set_xticks(x_inter)
    ax.set_xticklabels(x_ticks_blank)
    ax2.set_xticks(x_inter)
    ax2.set_xticklabels(x_ticks, rotation=80, fontsize=11)
    ax.grid(which='major', color='#D5D5D5', alpha=0.5)
    save_name = plt_title+'_'+notes+'a_1D-ALE_Plot.png'
    if directory is None:
        save_name = Path('', save_name)
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, save_name)
    plt.savefig(save_name)
    plt.close()

    o = np.hstack((np.array(x_idx).reshape(len(x_idx), 1), heter_value.reshape(heter_value.shape[0], 1)))
    save_name = plt_title+'_'+notes+'b_1D-ALE_Data_'+f_name+'.csv'
    if directory is None:
        save_name = Path('', save_name)
    else:
        save_name = Path('', directory, save_name)
    np.savetxt(save_name, o, fmt='%s', delimiter=',')


def ReduceDimScatter(X_new, y, plt_title, notes='', directory=None):
    """
    Draw 2D Scatter plot using 2D data from t-SNE or PCA

    :param X_new: Dimension-reduced X matrix, numpy.array, shape (num_samples, 2)
    :param y: Label values, numpy.array, shape (num_samples, )
    :param plt_title: Title of line plot, default as 'Reg', name of regressor is recommended
    :param notes: Notes in file name, default as ''
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'

    :return: None

    example:
    DChemPy.Drawing.ReduceDimScatter(X_new, y, plt_title='t-SNE', directory='xxx')
    """
    if type(y) is np.ndarray:
        y = y.reshape(y.shape[0], )
    elif type(y) is list:
        y = np.array(y).reshape(len(y), )
    save_name = plt_title+'_'+notes+'_ReduceDim_X.csv'
    if directory is None:
        save_name = Path('', save_name)
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, save_name)
    np.savetxt(save_name, X_new, fmt='%s', delimiter=',')
    fig = plt.figure(figsize=(10, 8), dpi=300)
    ax = fig.add_axes([0.11, 0.08, 0.88, 0.815])
    sc = ax.scatter(X_new[:, 0], X_new[:, 1], alpha=0.75, c=y, cmap='viridis', marker='o', zorder=20)
    ax.set_xlabel('Dim-1', fontsize=17)
    ax.set_ylabel('Dim-2', fontsize=17)
    plt.suptitle('Scatter of Dimesion Reducing Results by '+plt_title+'\n', fontsize=21)
    cb = plt.colorbar(sc)
    cb.set_label('Label Value', fontsize=17)
    plt.grid(which='major', color='#D5D5D5', alpha=0.5, zorder=1)
    save_name = plt_title+'_'+notes+'_ReduceDim-Scatter.png'
    if directory is None:
        save_name = Path('', save_name)
    else:
        save_name = Path('', directory, save_name)
    plt.savefig(save_name)
    plt.close()



"""
Modify dataset

Contents:
    DChemPy.Dataset.YCtoB: Change Continuous label values into Binary values {0, 1}
    DChemPy.Dataset.XCtoB: Change Continuous feature matrix into Binary features {0, 1}
    DChemPy.Dataset.ExtractX: Cut the Feature Matrix by given feature name list
    DChemPy.Dataset.DrawDisY: Draw the distribution of label values by hist plot
"""

import numpy as np
import math
import matplotlib.pyplot as plt
import os
from pathlib import Path


def YCtoB(y, directory=None, notes='01'):
    """
    Turn Continuous label values into Binary values {0, 1}

    :param y: Label Values with type of numpy.array and shape of (num_samples, 1)
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'
    :param notes: Notes in file name, default as '01'

    :return: None

    example:
    DChemPy.Dataset.YCtoB(y, directory='data_out')
    """
    if type(y) is np.ndarray:
        y = y.reshape(y.shape[0], 1)
    elif type(y) is list:
        y = np.array(y).reshape(len(y), 1)
    y_01 = np.zeros((y.shape[0], 1))
    y_mean = np.mean(y)
    for i in range(y.shape[0]):
        if y[i, 0] >= y_mean:
            y_01[i, 0] += 1
    if directory is None:
        np.savetxt('Values' + notes + '_' + str(y_01.shape[0]) + '.csv', y_01, fmt='%s', delimiter=',')
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, 'Values' + notes + '_' + str(y_01.shape[0]) + '.csv')
        np.savetxt(save_name, y_01, fmt='%s', delimiter=',')


def XCtoB(X, title, divide_pos=[-0.5, -0.15, 0, 0.15, 0.5], directory=None, notes='01'):
    """
    Turn Continuous X into Binary values {0, 1}
    noted: title will be expanded by the set of divide_pos

    :param X: Feature Matrix
    :param title: Title with type of numpy.array and shape of (num_featrues, 1)
    :param divide_pos: List of dividing points of feature values, default as [-0.5, -0.15, 0, 0.15, 0.5]
                       noted: One continuous feature will be divided into len(divide_pos)+1 binary features
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'
    :param notes: Notes in file name, default as '01'

    :return: None

    example:
    DChemPy.Dataset.YCtoB(y, directory='data_out')
    """
    if type(title) is np.ndarray:
        title = title.reshape(title.shape[0], 1)
    elif type(title) is list:
        title = np.array(title).reshape(len(title), 1)
    if directory is None:
        save_name = Path('', 'Interval_Data.txt')
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, 'Interval_Data.txt')
    inter = []
    f1 = open(save_name, 'w')
    for i in range(X.shape[1]):
        t_min = min(X[:, i])
        t_max = max(X[:, i]) * 1.01
        t_mean = np.mean(X[:, i])
        if 'Wavelength' not in title[i, 0] and 'AbsorbedEnergy' not in title[i, 0]:
            if t_min != 0:
                temp = [round(t_min, max(1, 2 - math.floor(np.log10(abs(t_min)))))]
            else:
                temp = [0.0]
            for j in range(len(divide_pos)):
                if divide_pos[j] < 0:
                    t = t_mean + divide_pos[j] * (t_mean - t_min)
                elif divide_pos[j] > 0:
                    t = t_mean + divide_pos[j] * (t_max - t_mean)
                else:
                    t = t_mean
                if t != 0:
                    temp.append(round(t, max(1, 2 - math.floor(np.log10(abs(t))))))
                else:
                    temp.append(0)
            if t_max != 0:
                temp.append(round(t_max, max(1, 2 - math.floor(np.log10(abs(t_max))))))
            else:
                temp.append(0)
        elif 'Wavelength' in title[i, 0]:
            t1 = math.ceil(t_min / 50)
            t2 = math.floor(t_max / 50)
            temp = np.arange(t1 * 50, t2 * 50, 50).flatten().tolist()
        else:
            temp = []
            for j in range(len(inter[-1])):
                temp.append(
                    round(1239.842 / inter[-1][j], max(1, 2 - math.floor(np.log10(abs(1239.842 / inter[-1][j]))))))
            temp = np.sort(temp)
        inter.append(temp)
        for j in range(len(temp)):
            f1.write(str(temp[j]))
            if j != len(temp) - 1:
                f1.write(',')
        f1.write('\n')
    f1.close()
    m = []
    for i in range(X.shape[1]):
        t = np.mean(X[:, i])
        s = str(round(t, max(1, 2 - math.floor(np.log10(abs(t))))))
        m.append(s)
    out_l = []
    title_l = []
    t_num = title.shape[0]
    for i in range(t_num):
        tt = title[i] + '(' + m[i] + ')'
        print(tt)
        d = inter[i]
        d = np.array(d)
        d = d.astype(str).flatten().tolist()
        print(d)
        d_temp = np.zeros((X.shape[0], len(d) + 1))
        t_list = X[:, i].flatten().tolist()
        for j in range(X.shape[0]):
            t = t_list[j]
            for k in range(len(d) + 1):
                if k == 0 and t < float(d[0]):
                    d_temp[j, 0] = 1
                elif k == len(d) and t >= float(d[-1]):
                    d_temp[j, -1] = 1
                elif (k != 0 and k != len(d)) and (float(d[k - 1]) <= t < float(d[k])):
                    d_temp[j, k] = 1
        out_l.append(d_temp)
        e = []
        for k in range(len(d) + 1):
            if k == 0:
                e.append(tt + '<' + d[0])
            elif k == len(d):
                e.append(d[-1] + '<=' + tt)
            else:
                e.append(d[k - 1] + '<=' + tt + '<' + d[k])
        e = np.array(e).reshape(len(e), 1)
        title_l.append(e)
    o = out_l[0].copy()
    for i in range(len(out_l) - 1):
        o = np.hstack((o, out_l[i + 1]))
    del_list = []
    for j in range(o.shape[1]):
        if np.sum(o[:, j]) == 0:
            del_list.append(j)
    o = np.delete(o, del_list, axis=1)
    ttt = title_l[0].copy()
    for i in range(len(title_l) - 1):
        ttt = np.vstack((ttt, title_l[i + 1]))
    ttt = np.delete(ttt, del_list, axis=0)
    X_temp = np.zeros((o.shape[0], o.shape[1]))
    index = -1
    a = []
    for i in range(o.shape[0]):
        if str(o[i, :].flatten().tolist()) not in a:
            a.append(str(o[i, :].flatten().tolist()))
            index += 1
            for j in range(o.shape[1]):
                X_temp[index, j] += o[i, j]
        else:
            ii = a.index(str(o[i, :].flatten().tolist()))
    X_out = X_temp[:index + 1, :]
    if directory is None:
        np.savetxt('Features_'+notes+str(X_out.shape[0])+'_'+str(X_out.shape[1])+'.csv', X_out, fmt='%s', delimiter=',')
        np.savetxt('Title_'+notes+str(X_out.shape[1])+'.csv', ttt, fmt='%s', delimiter=',')
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, 'Features_' + notes + str(X_out.shape[0]) + '_' + str(X_out.shape[1]) + '.csv')
        np.savetxt(save_name, X_out, fmt='%s', delimiter=',')
        save_name = Path('', directory, 'Title_' + notes + str(X_out.shape[1]) + '.csv')
        np.savetxt(save_name, ttt, fmt='%s', delimiter=',')


def ExtractX(X, origin_title, subset_title, directory=None, notes=''):
    """
    Cut the Feature Matrix by given feature name list

    :param X: Feature Matrix
    :param origin_title: old title with type of numpy.array and shape of (num_featrues, 1)
    :param subset_title: new title with type of numpy.array and shape of (num_featrues_new, 1)
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'
    :param notes: Notes in file name, default as ''

    :return: None

    example:
    DChemPy.Dataset.ExtractX(X, title_old, title_new, directory='data_out')
    """
    data_out = []
    title_out = []
    if type(origin_title) is np.ndarray:
        t_l_f = list(origin_title.flatten())
    if type(subset_title) is np.ndarray:
        t_l = list(subset_title.flatten())
    for i in range(len(t_l_f)):
        t = t_l_f[i]
        if t in t_l:
            title_out.append(t)
            data_out.append(i)
    X_out = X[:, data_out]
    title_out = np.array(title_out).reshape(len(title_out), 1)
    if directory is None:
        np.savetxt('Features_'+notes+str(X_out.shape[0])+'_'+str(X_out.shape[1])+'.csv', X_out, fmt='%s', delimiter=',')
        np.savetxt('Title_'+notes+str(title_out.shape[0])+'.csv', title_out, fmt='%s', delimiter=',')
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, 'Features_' + notes + str(X_out.shape[0]) + '_' + str(X_out.shape[1]) + '.csv')
        np.savetxt(save_name, X_out, fmt='%s', delimiter=',')
        save_name = Path('', directory, 'Title_' + notes + str(title_out.shape[0]) + '.csv')
        np.savetxt(save_name, title_out, fmt='%s', delimiter=',')


def DrawDisY(y, name='Y', bins=30, directory=None):
    """
    Draw the distribution of label values

    :param y: Label Value with type of numpy.array and shape of (num_samples, 1)
    :param name: Title of hist plot and Name of saved plot, default as 'Y'
    :param bins: number of bins in hist plot, default as 30
    :param directory: Directory to save outputs, default as 'None', for example 'data_out'

    :return: None

    example:
    DChemPy.Dataset.DrawDisY(y, directory='data_out')
    """
    plt.figure(figsize=(5, 3), dpi=250)
    plt.title('y_dis of ' + name, fontsize=15)
    plt.hist(y, bins=bins, facecolor='lightskyblue', edgecolor='dodgerblue')
    plt.ylabel('Counts', fontsize=15)
    plt.xlabel('Generation(Epoch)', fontsize=15)
    save_name = name + '.png'
    if directory is None:
        plt.savefig(save_name, bbox_inches='tight')
    else:
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_name = Path('', directory, save_name)
        plt.savefig(save_name, bbox_inches='tight')



"""
About DChemPy:
    Made by Yiheng Dai
    Last Update: 2022.11.21
    Most codes were completed during 2020-2022 in Xiamen University.
    Made for Machine Learning projects without deep learning in chemistry.

Contents:
    DChemPy.Chem: Process SMILES
    DChemPy.Dataset: Modify Dataset
    DChemPy.Drawing: Draw Plots in Machine Learning
    DChemPy.Featurization: Make SMILES into Feature Matrix
    DChemPy.Statistic: Basic Statistic Analysis

Original Paper:
    In press...
"""

from .Chem import *
from .Dataset import *
from .Drawing import *
from .Statistic import *
from .Featurization import *

VERSION = '1.1.1'
